@@ -22,16 +22,28 @@ you will screw up the layout where they are used in case expressions!
  * settings for the target plat instead). */
 #include "../includes/ghcautoconf.h"
 
-#if __GLASGOW_HASKELL__ >= 602
-#define SYSTEM_IO_ERROR System.IO.Error
-#else
-#define SYSTEM_IO_ERROR System.IO
-#endif
-
-#ifdef __GLASGOW_HASKELL__
+/* Global variables may not work in other Haskell implementations,
+ * but we need them currently! so the conditional on GLASGOW won't do. */
+#ifndef __HADDOCK__
+#if defined(__GLASGOW_HASKELL__) || !defined(__GLASGOW_HASKELL__)
 #define GLOBAL_VAR(name,value,ty)  \
-name = Util.global (value) :: IORef (ty); \
-{-# NOINLINE name #-}
+{-# NOINLINE name #-};             \
+name :: IORef (ty);                \
+name = Util.global (value);
+
+#define GLOBAL_MVAR(name,value,ty) \
+{-# NOINLINE name #-};             \
+name :: MVar (ty);                 \
+name = Util.globalMVar (value);
+#endif
+#else /* __HADDOCK__ */
+#define GLOBAL_VAR(name,value,ty)  \
+name :: IORef (ty);                \
+name = Util.global (value);
+
+#define GLOBAL_MVAR(name,value,ty) \
+name :: MVar (ty);                 \
+name = Util.globalMVar (value);
 #endif
 
 #define COMMA ,
@@ -39,6 +51,8 @@ name = Util.global (value) :: IORef (ty); \
 #ifdef DEBUG
 #define ASSERT(e) if (not (e)) then (assertPanic __FILE__ __LINE__) else
 #define ASSERT2(e,msg) if (not (e)) then (assertPprPanic __FILE__ __LINE__ (msg)) else
+#define MASSERT(e)      ASSERT(e) return ()
+#define MASSERT2(e,msg) ASSERT2(e,msg) return ()
 #define WARN( e, msg ) (warnPprTrace (e) __FILE__ __LINE__ (msg))
 #define ASSERTM(mbool) do { bool <- mbool; ASSERT(bool) return () }
 #define ASSERTM2(mbool,msg) do { bool <- mbool; ASSERT2(bool,msg) return () }
@@ -46,23 +60,15 @@ name = Util.global (value) :: IORef (ty); \
 -- We have to actually use all the variables we are given or we may get
 -- unused variable warnings when DEBUG is off.
 #define ASSERT(e)      if False && (not (e)) then panic "ASSERT" else
-#define ASSERT2(e,msg) if False && (not (e)) then pprPanic "ASSERT2" (msg) else
+#define ASSERT2(e,msg) if False && (const False (e,msg)) then pprPanic "ASSERT2" (msg) else
+#define MASSERT(e)      ASSERT(e) return ()
+#define MASSERT2(e,msg) ASSERT2(e,msg) return ()
 #define ASSERTM(e)       do { let { _mbool = (e) } }
 -- Here we deliberately don't use when as Control.Monad might not be imported
 #define ASSERTM2(e,msg)  do { let { _mbool = (e) }; if False then panic "ASSERTM2" else return () }
 #define WARN(e,msg)    if False && (e) then pprPanic "WARN" (msg) else
 #endif
 
--- This #ifndef lets us switch off the "import FastString"
--- when compiling FastString itself
-#ifndef COMPILING_FAST_STRING
--- 
-import qualified FastString as FS 
-#endif
-
-#define SLIT(x)	 (FS.mkLitString# (x#))
-#define FSLIT(x) (FS.mkFastString# (x#))
-
 -- Useful for declaring arguments to be strict
 #define STRICT1(f) f a | a `seq` False = undefined
 #define STRICT2(f) f a b | a `seq` b `seq` False = undefined
