@@ -1,15 +1,19 @@
-{-# OPTIONS -cpp #-}
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Distribution.ParseUtils
 -- Copyright   :  (c) The University of Glasgow 2004
--- 
--- Maintainer  :  libraries@haskell.org
--- Stability   :  alpha
+--
+-- Maintainer  :  cabal-devel@haskell.org
 -- Portability :  portable
 --
--- Utilities for parsing PackageDescription and InstalledPackageInfo.
-
+-- Utilities for parsing 'PackageDescription' and 'InstalledPackageInfo'.
+--
+-- The @.cabal@ file format is not trivial, especially with the introduction
+-- of configurations and the section syntax that goes with that. This module
+-- has a bunch of parsing functions that is used by the @.cabal@ parser and a
+-- couple others. It has the parsing framework code and also little parsers for
+-- many of the formats we get in various @.cabal@ file fields, like module
+-- names, comma separated lists etc.
 
 {- All rights reserved.
 
@@ -45,38 +49,44 @@ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. -}
 
 -- #hide
 module Distribution.ParseUtils (
-        LineNo, PError(..), PWarning, locatedErrorMsg, syntaxError, warning,
-	runP, ParseResult(..), catchParseError, parseFail,
-	Field(..), fName, lineNo,
-	FieldDescr(..), readFields,
-	parseFilePathQ, parseTokenQ,
-	parseModuleNameQ, parseDependency, parsePkgconfigDependency,
+        LineNo, PError(..), PWarning(..), locatedErrorMsg, syntaxError, warning,
+        runP, runE, ParseResult(..), catchParseError, parseFail, showPWarning,
+        Field(..), fName, lineNo,
+        FieldDescr(..), ppField, ppFields, readFields,
+        showFields, showSingleNamedField, parseFields,
+        parseFilePathQ, parseTokenQ, parseTokenQ',
+        parseModuleNameQ, parseBuildTool, parsePkgconfigDependency,
         parseOptVersion, parsePackageNameQ, parseVersionRangeQ,
-	parseTestedWithQ, parseLicenseQ, parseExtensionQ, 
-	parseSepList, parseCommaList, parseOptCommaList,
-	showFilePath, showToken, showTestedWith, showDependency, showFreeText,
-	field, simpleField, listField, commaListField, optsField, liftField,
-	parseReadS, parseReadSQ, parseQuoted,
+        parseTestedWithQ, parseLicenseQ, parseExtensionQ,
+        parseSepList, parseCommaList, parseOptCommaList,
+        showFilePath, showToken, showTestedWith, showFreeText, parseFreeText,
+        field, simpleField, listField, spaceListField, commaListField,
+        optsField, liftField, boolField, parseQuoted,
+
+        UnrecFieldParser, warnUnrec, ignoreUnrec,
   ) where
 
-import Distribution.Compiler (CompilerFlavor)
+import Distribution.Compiler (CompilerFlavor, parseCompilerFlavorCompat)
 import Distribution.License
 import Distribution.Version
-import Distribution.Package	( parsePackageName )
+         ( Version(..), VersionRange, anyVersion )
+import Distribution.Package     ( PackageName(..), Dependency(..) )
+import Distribution.ModuleName (ModuleName)
 import Distribution.Compat.ReadP as ReadP hiding (get)
+import Distribution.ReadE
+import Distribution.Text
+         ( Text(..) )
+import Distribution.Simple.Utils
+         ( intercalate, lowercase, normaliseLineEndings )
 import Language.Haskell.Extension (Extension)
 
 import Text.PrettyPrint.HughesPJ hiding (braces)
-import Data.Char        (isSpace, isUpper, toLower, isAlphaNum)
-import Data.Maybe	( fromMaybe)
-import Data.List        (intersperse)
-
-#ifdef DEBUG
-import Test.HUnit (Test(..), assertBool, Assertion, runTestTT, Counts, assertEqual)
-import IO
-import System.Environment ( getArgs )
-import Control.Monad ( zipWithM_ )
-#endif
+import Data.Char (isSpace, toLower, isAlphaNum, isDigit)
+import Data.Maybe       (fromMaybe)
+import Data.Tree as Tree (Tree(..), flatten)
+import qualified Data.Map as Map
+import Control.Monad (foldM)
+import System.FilePath (normalise)
 
 -- -----------------------------------------------------------------------------
 
@@ -88,18 +98,27 @@ data PError = AmbigousParse String LineNo
             | FromString String (Maybe LineNo)
         deriving Show
 
-type PWarning = String
+data PWarning = PWarning String
+              | UTFWarning LineNo String
+        deriving Show
+
+showPWarning :: FilePath -> PWarning -> String
+showPWarning fpath (PWarning msg) =
+  normalise fpath ++ ": " ++ msg
+showPWarning fpath (UTFWarning line fname) =
+  normalise fpath ++ ":" ++ show line
+        ++ ": Invalid UTF-8 text in the '" ++ fname ++ "' field."
 
 data ParseResult a = ParseFailed PError | ParseOk [PWarning] a
         deriving Show
 
 instance Monad ParseResult where
-	return x = ParseOk [] x
-	ParseFailed err >>= _ = ParseFailed err
-	ParseOk ws x >>= f = case f x of
-	                       ParseFailed err -> ParseFailed err
-			       ParseOk ws' x' -> ParseOk (ws'++ws) x'
-	fail s = ParseFailed (FromString s Nothing)
+        return x = ParseOk [] x
+        ParseFailed err >>= _ = ParseFailed err
+        ParseOk ws x >>= f = case f x of
+                               ParseFailed err -> ParseFailed err
+                               ParseOk ws' x' -> ParseOk (ws'++ws) x'
+        fail s = ParseFailed (FromString s Nothing)
 
 catchParseError :: ParseResult a -> (PError -> ParseResult a)
                 -> ParseResult a
@@ -112,17 +131,32 @@ parseFail = ParseFailed
 runP :: LineNo -> String -> ReadP a a -> String -> ParseResult a
 runP line fieldname p s =
   case [ x | (x,"") <- results ] of
-    [a] -> ParseOk [] a
+    [a] -> ParseOk (utf8Warnings line fieldname s) a
+    --TODO: what is this double parse thing all about?
+    --      Can't we just do the all isSpace test the first time?
     []  -> case [ x | (x,ys) <- results, all isSpace ys ] of
-             [a] -> ParseOk [] a
+             [a] -> ParseOk (utf8Warnings line fieldname s) a
              []  -> ParseFailed (NoParse fieldname line)
              _   -> ParseFailed (AmbigousParse fieldname line)
     _   -> ParseFailed (AmbigousParse fieldname line)
   where results = readP_to_S p s
 
+runE :: LineNo -> String -> ReadE a -> String -> ParseResult a
+runE line fieldname p s =
+    case runReadE p s of
+      Right a -> ParseOk (utf8Warnings line fieldname s) a
+      Left  e -> syntaxError line $
+        "Parse of field '" ++ fieldname ++ "' failed (" ++ e ++ "): " ++ s
+
+utf8Warnings :: LineNo -> String -> String -> [PWarning]
+utf8Warnings line fieldname s =
+  take 1 [ UTFWarning n fieldname
+         | (n,l) <- zip [line..] (lines s)
+         , '\xfffd' `elem` l ]
+
 locatedErrorMsg :: PError -> (Maybe LineNo, String)
-locatedErrorMsg (AmbigousParse f n) = (Just n, "Ambiguous parse in field '"++f++"'")
-locatedErrorMsg (NoParse f n)       = (Just n, "Parse of field '"++f++"' failed: ")
+locatedErrorMsg (AmbigousParse f n) = (Just n, "Ambiguous parse in field '"++f++"'.")
+locatedErrorMsg (NoParse f n)       = (Just n, "Parse of field '"++f++"' failed.")
 locatedErrorMsg (TabsError n)       = (Just n, "Tab used as indentation.")
 locatedErrorMsg (FromString s n)    = (n, s)
 
@@ -133,12 +167,12 @@ tabsError :: LineNo -> ParseResult a
 tabsError ln = ParseFailed $ TabsError ln
 
 warning :: String -> ParseResult ()
-warning s = ParseOk [s] ()
+warning s = ParseOk [PWarning s] ()
 
 -- | Field descriptor.  The parameter @a@ parameterizes over where the field's
 --   value is stored in.
-data FieldDescr a 
-  = FieldDescr 
+data FieldDescr a
+  = FieldDescr
       { fieldName     :: String
       , fieldGet      :: a -> Doc
       , fieldSet      :: LineNo -> String -> a -> ParseResult a
@@ -148,7 +182,7 @@ data FieldDescr a
       }
 
 field :: String -> (a -> Doc) -> (ReadP a a) -> FieldDescr a
-field name showF readF = 
+field name showF readF =
   FieldDescr name showF (\line val _st -> runP line name readF val)
 
 -- Lift a field descriptor storing into an 'a' to a field descriptor storing
@@ -156,9 +190,9 @@ field name showF readF =
 liftField :: (b -> a) -> (a -> b -> b) -> FieldDescr a -> FieldDescr b
 liftField get set (FieldDescr name showF parseF)
  = FieldDescr name (\b -> showF (get b))
-	(\line str b -> do
-	    a <- parseF line str (get b)
-	    return (set a b))
+        (\line str b -> do
+            a <- parseF line str (get b)
+            return (set a b))
 
 -- Parser combinator for simple fields.  Takes a field name, a pretty printer,
 -- a parser function, an accessor, and a setter, returns a FieldDescr over the
@@ -169,39 +203,118 @@ simpleField name showF readF get set
   = liftField get set $ field name showF readF
 
 commaListField :: String -> (a -> Doc) -> (ReadP [a] a)
-		 -> (b -> [a]) -> ([a] -> b -> b) -> FieldDescr b
-commaListField name showF readF get set = 
-  liftField get set $ 
+                 -> (b -> [a]) -> ([a] -> b -> b) -> FieldDescr b
+commaListField name showF readF get set =
+  liftField get set' $
     field name (fsep . punctuate comma . map showF) (parseCommaList readF)
+  where
+    set' xs b = set (get b ++ xs) b
+
+spaceListField :: String -> (a -> Doc) -> (ReadP [a] a)
+                 -> (b -> [a]) -> ([a] -> b -> b) -> FieldDescr b
+spaceListField name showF readF get set =
+  liftField get set' $
+    field name (fsep . map showF) (parseSpaceList readF)
+  where
+    set' xs b = set (get b ++ xs) b
 
 listField :: String -> (a -> Doc) -> (ReadP [a] a)
-		 -> (b -> [a]) -> ([a] -> b -> b) -> FieldDescr b
-listField name showF readF get set = 
-  liftField get set $ 
+                 -> (b -> [a]) -> ([a] -> b -> b) -> FieldDescr b
+listField name showF readF get set =
+  liftField get set' $
     field name (fsep . map showF) (parseOptCommaList readF)
+  where
+    set' xs b = set (get b ++ xs) b
 
 optsField :: String -> CompilerFlavor -> (b -> [(CompilerFlavor,[String])]) -> ([(CompilerFlavor,[String])] -> b -> b) -> FieldDescr b
-optsField name flavor get set = 
-   liftField (fromMaybe [] . lookup flavor . get) 
-	     (\opts b -> set (update flavor opts (get b)) b) $
-	field name (hsep . map text)
-		   (sepBy parseTokenQ' (munch1 isSpace))
+optsField name flavor get set =
+   liftField (fromMaybe [] . lookup flavor . get)
+             (\opts b -> set (update flavor opts (get b)) b) $
+        field name (hsep . map text)
+                   (sepBy parseTokenQ' (munch1 isSpace))
   where
         update f opts [] = [(f,opts)]
-	update f opts ((f',opts'):rest)
-           | f == f'   = (f, opts ++ opts') : rest
+        update f opts ((f',opts'):rest)
+           | f == f'   = (f, opts' ++ opts) : rest
            | otherwise = (f',opts') : update f opts rest
 
+-- TODO: this is a bit smelly hack. It's because we want to parse bool fields
+--       liberally but not accept new parses. We cannot do that with ReadP
+--       because it does not support warnings. We need a new parser framwork!
+boolField :: String -> (b -> Bool) -> (Bool -> b -> b) -> FieldDescr b
+boolField name get set = liftField get set (FieldDescr name showF readF)
+  where
+    showF = text . show
+    readF line str _
+      |  str == "True"  = ParseOk [] True
+      |  str == "False" = ParseOk [] False
+      | lstr == "true"  = ParseOk [caseWarning] True
+      | lstr == "false" = ParseOk [caseWarning] False
+      | otherwise       = ParseFailed (NoParse name line)
+      where
+        lstr = lowercase str
+        caseWarning = PWarning $
+          "The '" ++ name ++ "' field is case sensitive, use 'True' or 'False'."
+
+ppFields :: [FieldDescr a] -> a -> Doc
+ppFields fields x = vcat [ ppField name (getter x)
+                         | FieldDescr name getter _ <- fields]
+
+ppField :: String -> Doc -> Doc
+ppField name fielddoc = text name <> colon <+> fielddoc
+
+showFields :: [FieldDescr a] -> a -> String
+showFields fields = render . ($+$ text "") . ppFields fields
+
+showSingleNamedField :: [FieldDescr a] -> String -> Maybe (a -> String)
+showSingleNamedField fields f =
+  case [ get | (FieldDescr f' get _) <- fields, f' == f ] of
+    []      -> Nothing
+    (get:_) -> Just (render . ppField f . get)
+
+parseFields :: [FieldDescr a] -> a -> String -> ParseResult a
+parseFields fields initial = \str ->
+  readFields str >>= foldM setField initial
+  where
+    fieldMap = Map.fromList
+      [ (name, f) | f@(FieldDescr name _ _) <- fields ]
+    setField accum (F line name value) = case Map.lookup name fieldMap of
+      Just (FieldDescr _ _ set) -> set line value accum
+      Nothing -> do
+        warning ("Unrecognized field " ++ name ++ " on line " ++ show line)
+        return accum
+    setField accum f = do
+      warning ("Unrecognized stanza on line " ++ show (lineNo f))
+      return accum
+
+-- | The type of a function which, given a name-value pair of an
+--   unrecognized field, and the current structure being built,
+--   decides whether to incorporate the unrecognized field
+--   (by returning  Just x, where x is a possibly modified version
+--   of the structure being built), or not (by returning Nothing).
+type UnrecFieldParser a = (String,String) -> a -> Maybe a
+
+-- | A default unrecognized field parser which simply returns Nothing,
+--   i.e. ignores all unrecognized fields, so warnings will be generated.
+warnUnrec :: UnrecFieldParser a
+warnUnrec _ _ = Nothing
+
+-- | A default unrecognized field parser which silently (i.e. no
+--   warnings will be generated) ignores unrecognized fields, by
+--   returning the structure being built unmodified.
+ignoreUnrec :: UnrecFieldParser a
+ignoreUnrec _ x = Just x
+
 ------------------------------------------------------------------------------
 
--- The data type for our three syntactic categories 
-data Field 
+-- The data type for our three syntactic categories
+data Field
     = F LineNo String String
       -- ^ A regular @<property>: <value>@ field
     | Section LineNo String String [Field]
       -- ^ A section with a name and possible parameter.  The syntactic
       -- structure is:
-      -- 
+      --
       -- @
       --   <sectionname> <arg> {
       --     <field>*
@@ -231,16 +344,12 @@ fName (Section _ n _ _) = n
 fName _ = error "fname: not a field or section"
 
 readFields :: String -> ParseResult [Field]
-readFields input = 
-      ifelse
-  =<< mapM (mkField 0)
-  =<< mkTree (tokenise input)
+readFields input = ifelse
+               =<< mapM (mkField 0)
+               =<< mkTree tokens
 
-  where tokenise = concatMap tokeniseLine
-                 . trimLines
-                 . lines
-                 . normaliseLineEndings
-                 -- TODO: should decode UTF8
+  where ls = (lines . normaliseLineEndings) input
+        tokens = (concatMap tokeniseLine . trimLines) ls
 
 -- attach line number and determine indentation
 trimLines :: [String] -> [(LineNo, Indent, HasTabs, String)]
@@ -256,13 +365,13 @@ trimLines ls = [ (lineno, indent, hastabs, (trimTrailing l'))
 
 -- | We parse generically based on indent level and braces '{' '}'. To do that
 -- we split into lines and then '{' '}' tokens and other spans within a line.
-data Token = 
+data Token =
        -- | The 'Line' token is for bits that /start/ a line, eg:
        --
        -- > "\n  blah blah { blah"
        --
        -- tokenises to:
-       -- 
+       --
        -- > [Line n 2 False "blah blah", OpenBracket, Span n "blah"]
        --
        -- so lines are the only ones that can have nested layout, since they
@@ -284,11 +393,11 @@ data Token =
        --
        -- but that's not so common, people would normally use layout or
        -- brackets not both in a single @if else@ construct.
-       -- 
+       --
        -- > if ... { foo : bar }
        -- > else
        -- >    other
-       -- 
+       --
        -- this is ok
        Line LineNo Indent HasTabs String
      | Span LineNo                String  -- ^ span in a line, following brackets
@@ -308,7 +417,7 @@ tokeniseLine (n0, i, t, l) = case split n0 l of
           ("", '{' : s') ->             OpenBracket  n : split n s'
           (w , '{' : s') -> mkspan n w (OpenBracket  n : split n s')
           ("", '}' : s') ->             CloseBracket n : split n s'
-          (w , '}' : s') -> mkspan n w (CloseBracket n : split n s') 
+          (w , '}' : s') -> mkspan n w (CloseBracket n : split n s')
           (w ,        _) -> mkspan n w []
 
         mkspan n s ss | null s'   =             ss
@@ -320,13 +429,6 @@ trimLeading  = dropWhile isSpace
 trimTrailing = reverse . dropWhile isSpace . reverse
 
 
--- | Fix different systems silly line ending conventions
-normaliseLineEndings :: String -> String
-normaliseLineEndings [] = []
-normaliseLineEndings ('\r':'\n':s) = '\n' : normaliseLineEndings s -- windows
-normaliseLineEndings ('\r':s)      = '\n' : normaliseLineEndings s -- old osx
-normaliseLineEndings (  c :s)      =   c  : normaliseLineEndings s
-
 type SyntaxTree = Tree (LineNo, HasTabs, String)
 
 -- | Parse the stream of tokens into a tree of them, based on indent \/ layout
@@ -382,7 +484,7 @@ braces m a (Line n _ t l:OpenBracket n':ss) = do
     (sub, ss') <- braces n' [] ss
     braces m (Node (n,t,l) sub:a) ss'
 
-braces m a (Span n     l:OpenBracket n':ss) = do 
+braces m a (Span n     l:OpenBracket n':ss) = do
     (sub, ss') <- braces n' [] ss
     braces m (Node (n,False,l) sub:a) ss'
 
@@ -405,12 +507,12 @@ mkField d (Node (n,t,_) _) | d >= 1 && t = tabsError n
 mkField d (Node (n,_,l) ts) = case span (\c -> isAlphaNum c || c == '-') l of
   ([], _)       -> syntaxError n $ "unrecognised field or section: " ++ show l
   (name, rest)  -> case trimLeading rest of
-    (':':rest') -> do let followingLines = concatMap flatten ts
+    (':':rest') -> do let followingLines = concatMap Tree.flatten ts
                           tabs = not (null [()| (_,True,_) <- followingLines ])
                       if tabs && d >= 1
                         then tabsError n
                         else return $ F n (map toLower name)
-                                          (fieldValue rest' followingLines) 
+                                          (fieldValue rest' followingLines)
     rest'       -> do ts' <- mapM (mkField (d+1)) ts
                       return (Section n (map toLower name) rest' ts')
     where fieldValue firstLine followingLines =
@@ -418,7 +520,7 @@ mkField d (Node (n,_,l) ts) = case span (\c -> isAlphaNum c || c == '-') l of
                 followingLines' = map (\(_,_,s) -> stripDot s) followingLines
                 allLines | null firstLine' =              followingLines'
                          | otherwise       = firstLine' : followingLines'
-             in (concat . intersperse "\n") allLines
+             in intercalate "\n" allLines
           stripDot "." = ""
           stripDot s   = s
 
@@ -445,65 +547,69 @@ ifelse (Section n "else" _ _:_) = syntaxError n "stray 'else' with no preceding
 ifelse (Section n s a fs':fs) = do fs''  <- ifelse fs'
                                    fs''' <- ifelse fs
                                    return (Section n s a fs'' : fs''')
-ifelse (f:fs) = do fs' <- ifelse fs 
+ifelse (f:fs) = do fs' <- ifelse fs
                    return (f : fs')
 
 ------------------------------------------------------------------------------
 
 -- |parse a module name
-parseModuleNameQ :: ReadP r String
-parseModuleNameQ = parseQuoted modu <++ modu
- where modu = do 
-	  c <- satisfy isUpper
-	  cs <- munch (\x -> isAlphaNum x || x `elem` "_'.")
-	  return (c:cs)
+parseModuleNameQ :: ReadP r ModuleName
+parseModuleNameQ = parseQuoted parse <++ parse
 
 parseFilePathQ :: ReadP r FilePath
-parseFilePathQ = parseTokenQ 
+parseFilePathQ = parseTokenQ
   -- removed until normalise is no longer broken, was:
   --   liftM normalise parseTokenQ
 
-parseReadS :: Read a => ReadP r a
-parseReadS = readS_to_P reads
+parseBuildTool :: ReadP r Dependency
+parseBuildTool = do name <- parseBuildToolNameQ
+                    skipSpaces
+                    ver <- parseVersionRangeQ <++ return anyVersion
+                    skipSpaces
+                    return $ Dependency name ver
 
-parseDependency :: ReadP r Dependency
-parseDependency = do name <- parsePackageNameQ
-                     skipSpaces
-                     ver <- parseVersionRangeQ <++ return AnyVersion
-                     skipSpaces
-                     return $ Dependency name ver
+parseBuildToolNameQ :: ReadP r PackageName
+parseBuildToolNameQ = parseQuoted parseBuildToolName <++ parseBuildToolName
 
--- pkg-config allows versions and other letters in package names, 
+-- like parsePackageName but accepts symbols in components
+parseBuildToolName :: ReadP r PackageName
+parseBuildToolName = do ns <- sepBy1 component (ReadP.char '-')
+                        return (PackageName (intercalate "-" ns))
+  where component = do
+          cs <- munch1 (\c -> isAlphaNum c || c == '+' || c == '_')
+          if all isDigit cs then pfail else return cs
+
+-- pkg-config allows versions and other letters in package names,
 -- eg "gtk+-2.0" is a valid pkg-config package _name_.
 -- It then has a package version number like 2.10.13
 parsePkgconfigDependency :: ReadP r Dependency
 parsePkgconfigDependency = do name <- munch1 (\c -> isAlphaNum c || c `elem` "+-._")
                               skipSpaces
-                              ver <- parseVersionRangeQ <++ return AnyVersion
+                              ver <- parseVersionRangeQ <++ return anyVersion
                               skipSpaces
-                              return $ Dependency name ver
+                              return $ Dependency (PackageName name) ver
 
-parsePackageNameQ :: ReadP r String
-parsePackageNameQ = parseQuoted parsePackageName <++ parsePackageName 
+parsePackageNameQ :: ReadP r PackageName
+parsePackageNameQ = parseQuoted parse <++ parse
 
 parseVersionRangeQ :: ReadP r VersionRange
-parseVersionRangeQ = parseQuoted parseVersionRange <++ parseVersionRange
+parseVersionRangeQ = parseQuoted parse <++ parse
 
 parseOptVersion :: ReadP r Version
 parseOptVersion = parseQuoted ver <++ ver
-  where ver = parseVersion <++ return noVersion
-	noVersion = Version{ versionBranch=[], versionTags=[] }
+  where ver = parse <++ return noVersion
+        noVersion = Version{ versionBranch=[], versionTags=[] }
 
 parseTestedWithQ :: ReadP r (CompilerFlavor,VersionRange)
 parseTestedWithQ = parseQuoted tw <++ tw
-  where tw = do compiler <- parseReadS
-		skipSpaces
-		version <- parseVersionRange <++ return AnyVersion
-		skipSpaces
-		return (compiler,version)
+  where tw = do compiler <- parseCompilerFlavorCompat
+                skipSpaces
+                version <- parse <++ return anyVersion
+                skipSpaces
+                return (compiler,version)
 
 parseLicenseQ :: ReadP r License
-parseLicenseQ = parseQuoted parseReadS <++ parseReadS
+parseLicenseQ = parseQuoted parse <++ parse
 
 -- urgh, we can't define optQuotes :: ReadP r a -> ReadP r a
 -- because the "compat" version of ReadP isn't quite powerful enough.  In
@@ -511,24 +617,27 @@ parseLicenseQ = parseQuoted parseReadS <++ parseReadS
 -- Hence the trick above to make 'lic' polymorphic.
 
 parseExtensionQ :: ReadP r Extension
-parseExtensionQ = parseQuoted parseReadS <++ parseReadS
+parseExtensionQ = parseQuoted parse <++ parse
 
--- | Parse something optionally wrapped in quotes.
-parseReadSQ :: Read a => ReadP r a
-parseReadSQ = parseQuoted parseReadS <++ parseReadS
+parseHaskellString :: ReadP r String
+parseHaskellString = readS_to_P reads
 
 parseTokenQ :: ReadP r String
-parseTokenQ = parseReadS <++ munch1 (\x -> not (isSpace x) && x /= ',')
+parseTokenQ = parseHaskellString <++ munch1 (\x -> not (isSpace x) && x /= ',')
 
 parseTokenQ' :: ReadP r String
-parseTokenQ' = parseReadS <++ munch1 (\x -> not (isSpace x))
+parseTokenQ' = parseHaskellString <++ munch1 (\x -> not (isSpace x))
 
 parseSepList :: ReadP r b
-	     -> ReadP r a -- ^The parser for the stuff between commas
+             -> ReadP r a -- ^The parser for the stuff between commas
              -> ReadP r [a]
 parseSepList sepr p = sepBy p separator
     where separator = skipSpaces >> sepr >> skipSpaces
 
+parseSpaceList :: ReadP r a -- ^The parser for the stuff between commas
+               -> ReadP r [a]
+parseSpaceList p = sepBy p skipSpaces
+
 parseCommaList :: ReadP r a -- ^The parser for the stuff between commas
                -> ReadP r [a]
 parseCommaList = parseSepList (ReadP.char ',')
@@ -540,6 +649,9 @@ parseOptCommaList = parseSepList (optional (ReadP.char ','))
 parseQuoted :: ReadP r a -> ReadP r a
 parseQuoted p = between (ReadP.char '"') (ReadP.char '"') p
 
+parseFreeText :: ReadP.ReadP s String
+parseFreeText = ReadP.munch (const True)
+
 -- --------------------------------------------
 -- ** Pretty printing
 
@@ -554,181 +666,9 @@ showToken str
   where dodgy c = isSpace c || c == ','
 
 showTestedWith :: (CompilerFlavor,VersionRange) -> Doc
-showTestedWith (compiler,version) = text (show compiler ++ " " ++ showVersionRange version)
-
-showDependency :: Dependency -> Doc
-showDependency (Dependency name ver) = text name <+> text (showVersionRange ver)
+showTestedWith (compiler, version) = text (show compiler) <+> disp version
 
 -- | Pretty-print free-format text, ensuring that it is vertically aligned,
 -- and with blank lines replaced by dots for correct re-parsing.
 showFreeText :: String -> Doc
 showFreeText s = vcat [text (if null l then "." else l) | l <- lines s]
-
--- --------------------------------------------
--- ** Tree bits
-
--- Data.Tree was not present in ghc-6.2, and we only need these bits:
-
-data Tree a = Node a (Forest a)
-type Forest a = [Tree a]
-
-flatten :: Tree a -> [a]
-flatten t = squish t []
-  where squish (Node x ts) xs = x:foldr squish xs ts
-
-------------------------------------------------------------------------------
--- TESTING
-
-#ifdef DEBUG
-test_readFields = case 
-                    readFields testFile 
-                  of
-                    ParseOk _ x -> x == expectedResult
-                    _ -> False
-  where 
-    testFile = unlines $
-          [ "Cabal-version: 3"
-          , ""
-          , "Description: This is a test file   "
-          , "  with a description longer than two lines.  "
-          , "if os(windows) {"
-          , "  License:  You may not use this software"
-          , "    ."
-          , "    If you do use this software you will be seeked and destroyed."
-          , "}"
-          , "if os(linux) {"
-          , "  Main-is:  foo1  "
-          , "}"
-          , ""
-          , "if os(vista) {"
-          , "  executable RootKit {"
-          , "    Main-is: DRMManager.hs"
-          , "  }"
-          , "} else {"
-          , "  executable VistaRemoteAccess {"
-          , "    Main-is: VCtrl"
-          , "}}"
-          , ""
-          , "executable Foo-bar {"
-          , "  Main-is: Foo.hs"
-          , "}"
-          ]
-    expectedResult = 
-          [ F 1 "cabal-version" "3"
-          , F 3 "description" 
-                  "This is a test file\nwith a description longer than two lines."
-          , IfBlock 5 "os(windows) " 
-              [ F 6 "license" 
-                      "You may not use this software\n\nIf you do use this software you will be seeked and destroyed."
-              ]
-              []
-          , IfBlock 10 "os(linux) " 
-              [ F 11 "main-is" "foo1" ] 
-              [ ]
-          , IfBlock 14 "os(vista) " 
-              [ Section 15 "executable" "RootKit " 
-                [ F 16 "main-is" "DRMManager.hs"]
-              ] 
-              [ Section 19 "executable" "VistaRemoteAccess "
-                 [F 20 "main-is" "VCtrl"]
-              ]
-          , Section 23 "executable" "Foo-bar " 
-              [F 24 "main-is" "Foo.hs"]
-          ]
-
-test_readFieldsCompat' = case test_readFieldsCompat of
-                           ParseOk _ fs -> mapM_ (putStrLn . show) fs
-                           x -> putStrLn $ "Failed: " ++ show x
-test_readFieldsCompat = readFields testPkgDesc
-  where 
-    testPkgDesc = unlines [
-        "-- Required",
-        "Name: Cabal",
-        "Version: 0.1.1.1.1-rain",
-        "License: LGPL",
-        "License-File: foo",
-        "Copyright: Free Text String",
-        "Cabal-version: >1.1.1",
-        "-- Optional - may be in source?",
-        "Author: Happy Haskell Hacker",
-        "Homepage: http://www.haskell.org/foo",
-        "Package-url: http://www.haskell.org/foo",
-        "Synopsis: a nice package!",
-        "Description: a really nice package!",
-        "Category: tools",
-        "buildable: True",
-        "CC-OPTIONS: -g -o",
-        "LD-OPTIONS: -BStatic -dn",
-        "Frameworks: foo",
-        "Tested-with: GHC",
-        "Stability: Free Text String",
-        "Build-Depends: haskell-src, HUnit>=1.0.0-rain",
-        "Other-Modules: Distribution.Package, Distribution.Version,",
-        "                Distribution.Simple.GHCPackageConfig",
-        "Other-files: file1, file2",
-        "Extra-Tmp-Files:    file1, file2",
-        "C-Sources: not/even/rain.c, such/small/hands",
-        "HS-Source-Dirs: src, src2",
-        "Exposed-Modules: Distribution.Void, Foo.Bar",
-        "Extensions: OverlappingInstances, TypeSynonymInstances",
-        "Extra-Libraries: libfoo, bar, bang",
-        "Extra-Lib-Dirs: \"/usr/local/libs\"",
-        "Include-Dirs: your/slightest, look/will",
-        "Includes: /easily/unclose, /me, \"funky, path\\\\name\"",
-        "Install-Includes: /easily/unclose, /me, \"funky, path\\\\name\"",
-        "GHC-Options: -fTH -fglasgow-exts",
-        "Hugs-Options: +TH",
-        "Nhc-Options: ",
-        "Jhc-Options: ",
-        "",
-        "-- Next is an executable",
-        "Executable: somescript",
-        "Main-is: SomeFile.hs",
-        "Other-Modules: Foo1, Util, Main",
-        "HS-Source-Dir: scripts",
-        "Extensions: OverlappingInstances",
-        "GHC-Options: ",
-        "Hugs-Options: ",
-        "Nhc-Options: ",
-        "Jhc-Options: "
-        ]
-{-
-test' = do h <- openFile "../Cabal.cabal" ReadMode
-           s <- hGetContents h
-           let r = readFields s
-           case r of
-             ParseOk _ fs -> mapM_ (putStrLn . show) fs
-             x -> putStrLn $ "Failed: " ++ show x
-           putStrLn "==================="
-           mapM_ (putStrLn . show) $
-                 merge . zip [1..] . lines $ s
-           hClose h
--}
-
--- ghc -DDEBUG --make Distribution/ParseUtils.hs -o test
-
-main :: IO ()
-main = do
-  inputFiles <- getArgs
-  ok <- mapM checkResult inputFiles
-
-  zipWithM_ summary inputFiles ok
-  putStrLn $ show (length (filter not ok)) ++ " out of " ++ show (length ok) ++ " failed"
-
-  where summary f True  = return ()
-        summary f False = putStrLn $ f  ++ " failed :-("
-
-checkResult :: FilePath -> IO Bool
-checkResult inputFile = do
-  file <- readFile inputFile
-  case readFields file of
-    ParseOk _ result -> do
-       hPutStrLn stderr $ inputFile ++ " parses ok :-)"
-       return True
-    ParseFailed err -> do
-       hPutStrLn stderr $ inputFile ++ " parse failed:"
-       hPutStrLn stderr $ show err
-       return False
-
--- 
-#endif
