@@ -1,7 +1,4 @@
-{-# OPTIONS -cpp -fffi #-}
--- OPTIONS required for ghc-6.4.x compat, and must appear first
 {-# LANGUAGE CPP, ForeignFunctionInterface #-}
-{-# OPTIONS_GHC -cpp -fffi #-}
 {-# OPTIONS_NHC98 -cpp #-}
 {-# OPTIONS_JHC -fcpp -fffi #-}
 -----------------------------------------------------------------------------
@@ -59,6 +56,7 @@ module Distribution.Simple.InstallDirs (
         absoluteInstallDirs,
         CopyDest(..),
         prefixRelativeInstallDirs,
+        substituteInstallDirTemplates,
 
         PathTemplate,
         PathTemplateVariable(..),
@@ -66,7 +64,10 @@ module Distribution.Simple.InstallDirs (
         fromPathTemplate,
         substPathTemplate,
         initialPathTemplateEnv,
-        fullPathTemplateEnv,
+        platformTemplateEnv,
+        compilerTemplateEnv,
+        packageTemplateEnv,
+        installDirsTemplateEnv,
   ) where
 
 
@@ -82,7 +83,7 @@ import System.FilePath (dropDrive)
 import Distribution.Package
          ( PackageIdentifier, packageName, packageVersion )
 import Distribution.System
-         ( OS(..), buildOS, buildArch )
+         ( OS(..), buildOS, Platform(..), buildPlatform )
 import Distribution.Compiler
          ( CompilerId, CompilerFlavor(..) )
 import Distribution.Text
@@ -217,7 +218,8 @@ type InstallDirTemplates = InstallDirs PathTemplate
 defaultInstallDirs :: CompilerFlavor -> Bool -> Bool -> IO InstallDirTemplates
 defaultInstallDirs comp userInstall hasLibs = do
   windowsProgramFilesDir <- getWindowsProgramFilesDir
-  userInstallPrefix <- getAppUserDataDirectory "cabal"
+  userInstallPrefix      <- getAppUserDataDirectory "cabal"
+  lhcPrefix              <- getAppUserDataDirectory "lhc"
   return $ fmap toPathTemplate $ InstallDirs {
       prefix       = if userInstall
                        then userInstallPrefix
@@ -227,10 +229,13 @@ defaultInstallDirs comp userInstall hasLibs = do
       bindir       = "$prefix" </> "bin",
       libdir       = case buildOS of
         Windows   -> "$prefix"
-        _other    -> "$prefix" </> "lib",
+        _other    -> case comp of
+                       LHC | userInstall -> lhcPrefix
+                       _                 -> "$prefix" </> "lib",
       libsubdir    = case comp of
            Hugs   -> "hugs" </> "packages" </> "$pkg"
            JHC    -> "$compiler"
+           LHC    -> "$compiler"
            _other -> "$pkgid" </> "$compiler",
       dynlibdir    = "$libdir",
       libexecdir   = case buildOS of
@@ -266,9 +271,9 @@ defaultInstallDirs comp userInstall hasLibs = do
 -- 'PathTemplate's that still have the 'PrefixVar' in them. Doing this makes it
 -- each to check which paths are relative to the $prefix.
 --
-substituteTemplates :: PackageIdentifier -> CompilerId
-                    -> InstallDirTemplates -> InstallDirTemplates
-substituteTemplates pkgId compilerId dirs = dirs'
+substituteInstallDirTemplates :: PathTemplateEnv
+                              -> InstallDirTemplates -> InstallDirTemplates
+substituteInstallDirTemplates env dirs = dirs'
   where
     dirs' = InstallDirs {
       -- So this specifies exactly which vars are allowed in each template
@@ -288,8 +293,6 @@ substituteTemplates pkgId compilerId dirs = dirs'
       haddockdir = subst haddockdir (prefixBinLibDataVars ++
                                       [docdirVar, htmldirVar])
     }
-    -- The initial environment has all the static stuff but no paths
-    env = initialPathTemplateEnv pkgId compilerId
     subst dir env' = substPathTemplate (env'++env) (dir dirs)
 
     prefixVar        = (PrefixVar,     prefix     dirs')
@@ -307,19 +310,23 @@ substituteTemplates pkgId compilerId dirs = dirs'
 -- substituting for all the variables in the abstract paths, to get real
 -- absolute path.
 absoluteInstallDirs :: PackageIdentifier -> CompilerId -> CopyDest
-                    -> InstallDirTemplates -> InstallDirs FilePath
+                    -> InstallDirs PathTemplate
+                    -> InstallDirs FilePath
 absoluteInstallDirs pkgId compilerId copydest dirs =
     (case copydest of
        CopyTo destdir -> fmap ((destdir </>) . dropDrive)
        _              -> id)
   . appendSubdirs (</>)
   . fmap fromPathTemplate
-  $ substituteTemplates pkgId compilerId dirs {
+  $ substituteInstallDirTemplates env dirs {
       prefix = case copydest of
         -- possibly override the prefix
         CopyPrefix p -> toPathTemplate p
         _            -> prefix dirs
     }
+  where
+    env = initialPathTemplateEnv pkgId compilerId
+
 
 -- |The location prefix for the /copy/ command.
 data CopyDest
@@ -343,10 +350,12 @@ prefixRelativeInstallDirs pkgId compilerId dirs =
   $ -- substitute the path template into each other, except that we map
     -- \$prefix back to $prefix. We're trying to end up with templates that
     -- mention no vars except $prefix.
-    substituteTemplates pkgId compilerId dirs {
+    substituteInstallDirTemplates env dirs {
       prefix = PathTemplate [Variable PrefixVar]
     }
   where
+    env = initialPathTemplateEnv pkgId compilerId
+
     -- If it starts with $prefix then it's relative and produce the relative
     -- path by stripping off $prefix/ or $prefix
     relative dir = case dir of
@@ -386,6 +395,8 @@ data PathTemplateVariable =
      | ExecutableNameVar -- ^ The executable name; used in shell wrappers
   deriving Eq
 
+type PathTemplateEnv = [(PathTemplateVariable, PathTemplate)]
+
 -- | Convert a 'FilePath' to a 'PathTemplate' including any template vars.
 --
 toPathTemplate :: FilePath -> PathTemplate
@@ -400,8 +411,7 @@ combinePathTemplate :: PathTemplate -> PathTemplate -> PathTemplate
 combinePathTemplate (PathTemplate t1) (PathTemplate t2) =
   PathTemplate (t1 ++ [Ordinary [pathSeparator]] ++ t2)
 
-substPathTemplate :: [(PathTemplateVariable, PathTemplate)]
-                  -> PathTemplate -> PathTemplate
+substPathTemplate :: PathTemplateEnv -> PathTemplate -> PathTemplate
 substPathTemplate environment (PathTemplate template) =
     PathTemplate (concatMap subst template)
 
@@ -411,38 +421,45 @@ substPathTemplate environment (PathTemplate template) =
                   Just (PathTemplate components) -> components
                   Nothing                        -> [component]
 
+
 -- | The initial environment has all the static stuff but no paths
-initialPathTemplateEnv :: PackageIdentifier -> CompilerId
-                       -> [(PathTemplateVariable, PathTemplate)]
+initialPathTemplateEnv :: PackageIdentifier -> CompilerId -> PathTemplateEnv
 initialPathTemplateEnv pkgId compilerId =
-  map (\(v,s) -> (v, PathTemplate [Ordinary s]))
-  [(PkgNameVar,  display (packageName pkgId))
-  ,(PkgVerVar,   display (packageVersion pkgId))
-  ,(PkgIdVar,    display pkgId)
-  ,(CompilerVar, display compilerId)
-  ,(OSVar,       display buildOS)    --these should be params if we want to be
-  ,(ArchVar,     display buildArch)  --able to do cross-platform configuation
+     packageTemplateEnv  pkgId
+  ++ compilerTemplateEnv compilerId
+  ++ platformTemplateEnv buildPlatform -- platform should be param if we want
+                                       -- to do cross-platform configuation
+
+packageTemplateEnv :: PackageIdentifier -> PathTemplateEnv
+packageTemplateEnv pkgId =
+  [(PkgNameVar,  PathTemplate [Ordinary $ display (packageName pkgId)])
+  ,(PkgVerVar,   PathTemplate [Ordinary $ display (packageVersion pkgId)])
+  ,(PkgIdVar,    PathTemplate [Ordinary $ display pkgId])
+  ]
+
+compilerTemplateEnv :: CompilerId -> PathTemplateEnv
+compilerTemplateEnv compilerId =
+  [(CompilerVar, PathTemplate [Ordinary $ display compilerId])
+  ]
+
+platformTemplateEnv :: Platform -> PathTemplateEnv
+platformTemplateEnv (Platform os arch) =
+  [(OSVar,       PathTemplate [Ordinary $ display os])
+  ,(ArchVar,     PathTemplate [Ordinary $ display arch])
+  ]
+
+installDirsTemplateEnv :: InstallDirs PathTemplate -> PathTemplateEnv
+installDirsTemplateEnv dirs =
+  [(PrefixVar,     prefix     dirs)
+  ,(BindirVar,     bindir     dirs)
+  ,(LibdirVar,     libdir     dirs)
+  ,(LibsubdirVar,  libsubdir  dirs)
+  ,(DatadirVar,    datadir    dirs)
+  ,(DatasubdirVar, datasubdir dirs)
+  ,(DocdirVar,     docdir     dirs)
+  ,(HtmldirVar,    htmldir    dirs)
   ]
 
-fullPathTemplateEnv :: PackageIdentifier -> CompilerId
-                    -> InstallDirs FilePath
-                    -> [(PathTemplateVariable, PathTemplate)]
-fullPathTemplateEnv pkgId compilerId dirs = env ++ dirEnv
-    where -- The initial environment has all the static stuff but no paths
-          env = initialPathTemplateEnv pkgId compilerId
-          -- And here are all the paths
-          dirEnv = [(PrefixVar,     toPathTemplate $ prefix     dirs),
-                    (BindirVar,     toPathTemplate $ bindir     dirs),
-                    (LibdirVar,     toPathTemplate $ libdir     dirs),
-                    -- This isn't defined in an InstallDirs FilePath
-                    -- as its value has already been appended to libdir:
-                    -- (LibsubdirVar,  toPathTemplate $ libsubdir  dirs),
-                    (DatadirVar,    toPathTemplate $ datadir    dirs),
-                    -- This isn't defined in an InstallDirs FilePath
-                    -- as its value has already been appended to datadir:
-                    -- (DatasubdirVar, toPathTemplate $ datasubdir dirs),
-                    (DocdirVar,     toPathTemplate $ docdir     dirs),
-                    (HtmldirVar,    toPathTemplate $ htmldir    dirs)]
 
 -- ---------------------------------------------------------------------------
 -- Parsing and showing path templates:
