@@ -42,19 +42,19 @@ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. -}
 
 module Distribution.Simple.NHC
   ( configure
-  , build
+  , buildLib, buildExe
   , installLib, installExe
   ) where
 
 import Distribution.Package
         ( PackageIdentifier, packageName, Package(..) )
 import Distribution.PackageDescription
-        ( PackageDescription(..), BuildInfo(..), Library(..), Executable(..),
-          withLib, withExe, hcOptions )
+        ( PackageDescription(..), BuildInfo(..), Library(..), Executable(..)
+        , hcOptions )
 import Distribution.ModuleName (ModuleName)
 import qualified Distribution.ModuleName as ModuleName
 import Distribution.Simple.LocalBuildInfo
-        ( LocalBuildInfo(..) )
+        ( LocalBuildInfo(..), ComponentLocalBuildInfo(..) )
 import Distribution.Simple.BuildPaths
         ( mkLibName, objExtension, exeExtension )
 import Distribution.Simple.Compiler
@@ -63,15 +63,16 @@ import Distribution.Simple.Compiler
 import Language.Haskell.Extension
         ( Extension(..) )
 import Distribution.Simple.Program
-        ( ProgramConfiguration, userMaybeSpecifyPath, requireProgram,
-          lookupProgram, ConfiguredProgram(programVersion), programPath,
-          nhcProgram, hmakeProgram, ldProgram, arProgram,
-          rawSystemProgramConf )
+         ( ProgramConfiguration, userMaybeSpecifyPath, programPath
+         , requireProgram, requireProgramVersion, lookupProgram
+         , nhcProgram, hmakeProgram, ldProgram, arProgram
+         , rawSystemProgramConf )
 import Distribution.Simple.Utils
-        ( die, info, findFileWithExtension,
-          createDirectoryIfMissingVerbose, copyFileVerbose, smartCopySources )
+        ( die, info, findFileWithExtension, findModuleFiles
+        , installOrdinaryFile, installExecutableFile, installOrdinaryFiles
+        , createDirectoryIfMissingVerbose )
 import Distribution.Version
-        ( Version(..), VersionRange(..), orLaterVersion )
+        ( Version(..), orLaterVersion )
 import Distribution.Verbosity
 import Distribution.Text
         ( display )
@@ -92,21 +93,22 @@ configure :: Verbosity -> Maybe FilePath -> Maybe FilePath
           -> ProgramConfiguration -> IO (Compiler, ProgramConfiguration)
 configure verbosity hcPath _hcPkgPath conf = do
 
-  (nhcProg, conf') <- requireProgram verbosity nhcProgram
-                          (orLaterVersion (Version [1,20] []))
-                          (userMaybeSpecifyPath "nhc98" hcPath conf)
-  let Just nhcVersion = programVersion nhcProg
+  (_nhcProg, nhcVersion, conf') <-
+    requireProgramVersion verbosity nhcProgram
+      (orLaterVersion (Version [1,20] []))
+      (userMaybeSpecifyPath "nhc98" hcPath conf)
 
-  (_hmakeProg, conf'') <- requireProgram verbosity hmakeProgram
-                          (orLaterVersion (Version [3,13] [])) conf'
-  (_ldProg, conf''')   <- requireProgram verbosity ldProgram AnyVersion conf''
-  (_arProg, conf'''')  <- requireProgram verbosity arProgram AnyVersion conf'''
+  (_hmakeProg, _hmakeVersion, conf'') <-
+    requireProgramVersion verbosity hmakeProgram
+     (orLaterVersion (Version [3,13] [])) conf'
+  (_ldProg, conf''')   <- requireProgram verbosity ldProgram conf''
+  (_arProg, conf'''')  <- requireProgram verbosity arProgram conf'''
 
   --TODO: put this stuff in a monad so we can say just:
   -- requireProgram hmakeProgram (orLaterVersion (Version [3,13] []))
-  -- requireProgram ldProgram AnyVersion
-  -- requireProgram ldPrograrProgramam AnyVersion
-  -- unless (null (cSources bi)) $ requireProgram ccProgram AnyVersion
+  -- requireProgram ldProgram anyVersion
+  -- requireProgram ldPrograrProgramam anyVersion
+  -- unless (null (cSources bi)) $ requireProgram ccProgram anyVersion
 
   let comp = Compiler {
         compilerId         = CompilerId NHC nhcVersion,
@@ -132,91 +134,96 @@ nhcLanguageExtensions =
 
 -- |FIX: For now, the target must contain a main module.  Not used
 -- ATM. Re-add later.
-build :: PackageDescription -> LocalBuildInfo -> Verbosity -> IO ()
-build pkg_descr lbi verbosity = do
+buildLib :: Verbosity -> PackageDescription -> LocalBuildInfo
+                      -> Library            -> ComponentLocalBuildInfo -> IO ()
+buildLib verbosity pkg_descr lbi lib clbi = do
   let conf = withPrograms lbi
       Just nhcProg = lookupProgram nhcProgram conf
-  withLib pkg_descr () $ \lib -> do
-    let bi = libBuildInfo lib
-        modules = exposedModules lib ++ otherModules bi
-        -- Unsupported extensions have already been checked by configure
-        extensionFlags = extensionsToFlags (compiler lbi) (extensions bi)
-    inFiles <- getModulePaths lbi bi modules
-    let targetDir = buildDir lbi
-        srcDirs  = nub (map takeDirectory inFiles)
-        destDirs = map (targetDir </>) srcDirs
-    mapM_ (createDirectoryIfMissingVerbose verbosity True) destDirs
-    rawSystemProgramConf verbosity hmakeProgram conf $
-         ["-hc=" ++ programPath nhcProg]
-      ++ nhcVerbosityOptions verbosity
-      ++ ["-d", targetDir, "-hidir", targetDir]
-      ++ extensionFlags
-      ++ maybe [] (hcOptions NHC . libBuildInfo)
-                             (library pkg_descr)
-      ++ concat [ ["-package", display (packageName pkg) ]
-                | pkg <- packageDeps lbi ]
-      ++ inFiles
+  let bi = libBuildInfo lib
+      modules = exposedModules lib ++ otherModules bi
+      -- Unsupported extensions have already been checked by configure
+      extensionFlags = extensionsToFlags (compiler lbi) (extensions bi)
+  inFiles <- getModulePaths lbi bi modules
+  let targetDir = buildDir lbi
+      srcDirs  = nub (map takeDirectory inFiles)
+      destDirs = map (targetDir </>) srcDirs
+  mapM_ (createDirectoryIfMissingVerbose verbosity True) destDirs
+  rawSystemProgramConf verbosity hmakeProgram conf $
+       ["-hc=" ++ programPath nhcProg]
+    ++ nhcVerbosityOptions verbosity
+    ++ ["-d", targetDir, "-hidir", targetDir]
+    ++ extensionFlags
+    ++ maybe [] (hcOptions NHC . libBuildInfo)
+                           (library pkg_descr)
+    ++ concat [ ["-package", display (packageName pkgid) ]
+              | (_, pkgid) <- componentPackageDeps clbi ]
+    ++ inFiles
 {-
-    -- build any C sources
-    unless (null (cSources bi)) $ do
-       info verbosity "Building C Sources..."
-       let commonCcArgs = (if verbosity >= deafening then ["-v"] else [])
-                       ++ ["-I" ++ dir | dir <- includeDirs bi]
-                       ++ [opt | opt <- ccOptions bi]
-                       ++ (if withOptimization lbi then ["-O2"] else [])
-       flip mapM_ (cSources bi) $ \cfile -> do
-         let ofile = targetDir </> cfile `replaceExtension` objExtension
-         createDirectoryIfMissingVerbose verbosity True (takeDirectory ofile)
-         rawSystemProgramConf verbosity hmakeProgram conf
-           (commonCcArgs ++ ["-c", cfile, "-o", ofile])
+  -- build any C sources
+  unless (null (cSources bi)) $ do
+     info verbosity "Building C Sources..."
+     let commonCcArgs = (if verbosity >= deafening then ["-v"] else [])
+                     ++ ["-I" ++ dir | dir <- includeDirs bi]
+                     ++ [opt | opt <- ccOptions bi]
+                     ++ (if withOptimization lbi then ["-O2"] else [])
+     flip mapM_ (cSources bi) $ \cfile -> do
+       let ofile = targetDir </> cfile `replaceExtension` objExtension
+       createDirectoryIfMissingVerbose verbosity True (takeDirectory ofile)
+       rawSystemProgramConf verbosity hmakeProgram conf
+         (commonCcArgs ++ ["-c", cfile, "-o", ofile])
 -}
-    -- link:
-    info verbosity "Linking..."
-    let --cObjs = [ targetDir </> cFile `replaceExtension` objExtension
-        --        | cFile <- cSources bi ]
-        libFilePath = targetDir </> mkLibName (packageId pkg_descr)
-        hObjs = [ targetDir </> ModuleName.toFilePath m <.> objExtension
-                | m <- modules ]
+  -- link:
+  info verbosity "Linking..."
+  let --cObjs = [ targetDir </> cFile `replaceExtension` objExtension
+      --        | cFile <- cSources bi ]
+      libFilePath = targetDir </> mkLibName (packageId pkg_descr)
+      hObjs = [ targetDir </> ModuleName.toFilePath m <.> objExtension
+              | m <- modules ]
 
-    unless (null hObjs {-&& null cObjs-}) $ do
-      -- first remove library if it exists
-      removeFile libFilePath `catchIO` \_ -> return ()
+  unless (null hObjs {-&& null cObjs-}) $ do
+    -- first remove library if it exists
+    removeFile libFilePath `catchIO` \_ -> return ()
 
-      let arVerbosity | verbosity >= deafening = "v"
-                      | verbosity >= normal = ""
-                      | otherwise = "c"
+    let arVerbosity | verbosity >= deafening = "v"
+                    | verbosity >= normal = ""
+                    | otherwise = "c"
 
-      rawSystemProgramConf verbosity arProgram (withPrograms lbi) $
-           ["q"++ arVerbosity, libFilePath]
-        ++ hObjs
---        ++ cObjs
+    rawSystemProgramConf verbosity arProgram (withPrograms lbi) $
+         ["q"++ arVerbosity, libFilePath]
+      ++ hObjs
+--    ++ cObjs
 
-  withExe pkg_descr $ \exe -> do
-    when (dropExtension (modulePath exe) /= exeName exe) $
-      die $ "hmake does not support exe names that do not match the name of "
-         ++ "the 'main-is' file. You will have to rename your executable to "
-         ++ show (dropExtension (modulePath exe))
-    let bi = buildInfo exe
-        modules = otherModules bi
-        -- Unsupported extensions have already been checked by configure
-        extensionFlags = extensionsToFlags (compiler lbi) (extensions bi)
-    inFiles <- getModulePaths lbi bi modules
-    let targetDir = buildDir lbi </> exeName exe
-        exeDir    = targetDir </> (exeName exe ++ "-tmp")
-        srcDirs   = nub (map takeDirectory (modulePath exe : inFiles))
-        destDirs  = map (exeDir </>) srcDirs
-    mapM_ (createDirectoryIfMissingVerbose verbosity True) destDirs
-    rawSystemProgramConf verbosity hmakeProgram conf $
-         ["-hc=" ++ programPath nhcProg]
-      ++ nhcVerbosityOptions verbosity
-      ++ ["-d", targetDir, "-hidir", targetDir]
-      ++ extensionFlags
-      ++ maybe [] (hcOptions NHC . libBuildInfo)
-                             (library pkg_descr)
-      ++ concat [ ["-package", display (packageName pkg) ]
-                | pkg <- packageDeps lbi ]
-      ++ inFiles
-      ++ [exeName exe]
+-- | Building an executable for NHC.
+buildExe :: Verbosity -> PackageDescription -> LocalBuildInfo
+                      -> Executable         -> ComponentLocalBuildInfo -> IO ()
+buildExe verbosity pkg_descr lbi exe clbi = do
+  let conf = withPrograms lbi
+      Just nhcProg = lookupProgram nhcProgram conf
+  when (dropExtension (modulePath exe) /= exeName exe) $
+    die $ "hmake does not support exe names that do not match the name of "
+       ++ "the 'main-is' file. You will have to rename your executable to "
+       ++ show (dropExtension (modulePath exe))
+  let bi = buildInfo exe
+      modules = otherModules bi
+      -- Unsupported extensions have already been checked by configure
+      extensionFlags = extensionsToFlags (compiler lbi) (extensions bi)
+  inFiles <- getModulePaths lbi bi modules
+  let targetDir = buildDir lbi </> exeName exe
+      exeDir    = targetDir </> (exeName exe ++ "-tmp")
+      srcDirs   = nub (map takeDirectory (modulePath exe : inFiles))
+      destDirs  = map (exeDir </>) srcDirs
+  mapM_ (createDirectoryIfMissingVerbose verbosity True) destDirs
+  rawSystemProgramConf verbosity hmakeProgram conf $
+       ["-hc=" ++ programPath nhcProg]
+    ++ nhcVerbosityOptions verbosity
+    ++ ["-d", targetDir, "-hidir", targetDir]
+    ++ extensionFlags
+    ++ maybe [] (hcOptions NHC . libBuildInfo)
+                           (library pkg_descr)
+    ++ concat [ ["-package", display (packageName pkgid) ]
+              | (_, pkgid) <- componentPackageDeps clbi ]
+    ++ inFiles
+    ++ [exeName exe]
 
 nhcVerbosityOptions :: Verbosity -> [String]
 nhcVerbosityOptions verbosity
@@ -235,7 +242,7 @@ getModulePaths lbi bi modules = sequence
 -- -----------------------------------------------------------------------------
 -- Installing
 
--- |Install executables for GHC.
+-- |Install executables for NHC.
 installExe :: Verbosity -- ^verbosity
            -> FilePath  -- ^install location
            -> FilePath  -- ^Build location
@@ -247,8 +254,9 @@ installExe verbosity pref buildPref (progprefix,progsuffix) exe
          let exeBaseName = exeName exe
              exeFileName = exeBaseName <.> exeExtension
              fixedExeFileName = (progprefix ++ exeBaseName ++ progsuffix) <.> exeExtension
-         copyFileVerbose verbosity (buildPref </> exeBaseName </> exeFileName)
-                                   (pref </> fixedExeFileName)
+         installExecutableFile verbosity
+           (buildPref </> exeBaseName </> exeFileName)
+           (pref </> fixedExeFileName)
 
 -- |Install for nhc98: .hi and .a files
 installLib    :: Verbosity -- ^verbosity
@@ -260,6 +268,7 @@ installLib    :: Verbosity -- ^verbosity
 installLib verbosity pref buildPref pkgid lib
     = do let bi = libBuildInfo lib
              modules = exposedModules lib ++ otherModules bi
-         smartCopySources verbosity [buildPref] pref modules ["hi"]
+         findModuleFiles [buildPref] ["hi"] modules
+           >>= installOrdinaryFiles verbosity pref
          let libName = mkLibName pkgid
-         copyFileVerbose verbosity (buildPref </> libName) (pref </> libName)
+         installOrdinaryFile verbosity (buildPref </> libName) (pref </> libName)
