@@ -20,7 +20,6 @@ module System.Win32.Types
 import Data.Maybe
 import Foreign
 import Foreign.C
-import Numeric (showHex)
 import Control.Exception
 import System.IO.Error
 import Data.Char
@@ -61,6 +60,8 @@ type SIZE_T        = DWORD
 
 type MbATOM        = Maybe ATOM
 
+type HRESULT       = LONG
+
 ----------------------------------------------------------------
 -- Pointers
 ----------------------------------------------------------------
@@ -204,7 +205,8 @@ failWith :: String -> ErrCode -> IO a
 failWith fn_name err_code = do
   c_msg <- getErrorMessage err_code
   msg <- peekTString c_msg
-  localFree c_msg
+  -- We ignore failure of freeing c_msg, given we're already failing
+  _ <- localFree c_msg
   c_maperrno -- turn GetLastError() into errno, which errnoToIOError knows
              -- how to convert to an IOException we can throw.
              -- XXX we should really do this directly.
