@@ -1,10 +1,12 @@
-{-# OPTIONS_GHC -cpp -fglasgow-exts #-}
-
--- #prune
+{-# LANGUAGE CPP #-}
+-- We cannot actually specify all the language pragmas, see ghc ticket #
+-- If we could, these are what they would be:
+{- LANGUAGE MagicHash, UnboxedTuples -}
+{-# OPTIONS_HADDOCK prune #-}
 
 -- |
 -- Module      : Data.ByteString.Char8
--- Copyright   : (c) Don Stewart 2006
+-- Copyright   : (c) Don Stewart 2006-2008
 -- License     : BSD-style
 --
 -- Maintainer  : dons@cse.unsw.edu.au
@@ -32,6 +34,11 @@
 --
 -- > import qualified Data.ByteString.Char8 as B
 --
+-- The Char8 interface to bytestrings provides an instance of IsString
+-- for the ByteString type, enabling you to use string literals, and
+-- have them implicitly packed to ByteStrings. Use -XOverloadedStrings
+-- to enable this.
+--
 
 module Data.ByteString.Char8 (
 
@@ -92,7 +99,6 @@ module Data.ByteString.Char8 (
         -- ** Accumulating maps
         mapAccumL,              -- :: (acc -> Char -> (acc, Char)) -> acc -> ByteString -> (acc, ByteString)
         mapAccumR,              -- :: (acc -> Char -> (acc, Char)) -> acc -> ByteString -> (acc, ByteString)
-        mapIndexed,             -- :: (Int -> Char -> Char) -> ByteString -> ByteString
 
         -- ** Generating and unfolding ByteStrings
         replicate,              -- :: Int -> Char -> ByteString
@@ -130,9 +136,9 @@ module Data.ByteString.Char8 (
         isPrefixOf,             -- :: ByteString -> ByteString -> Bool
         isSuffixOf,             -- :: ByteString -> ByteString -> Bool
         isInfixOf,              -- :: ByteString -> ByteString -> Bool
-        isSubstringOf,          -- :: ByteString -> ByteString -> Bool
 
         -- ** Search for arbitrary substrings
+        breakSubstring,         -- :: ByteString -> ByteString -> (ByteString,ByteString)
         findSubstring,          -- :: ByteString -> ByteString -> Maybe Int
         findSubstrings,         -- :: ByteString -> ByteString -> [Int]
 
@@ -205,9 +211,6 @@ module Data.ByteString.Char8 (
         hPutStr,                -- :: Handle -> ByteString -> IO ()
         hPutStrLn,              -- :: Handle -> ByteString -> IO ()
 
-        -- undocumented deprecated things:
-        join                    -- :: ByteString -> [ByteString] -> ByteString
-
   ) where
 
 import qualified Prelude as P
@@ -230,8 +233,8 @@ import qualified Data.ByteString.Unsafe as B
 import Data.ByteString (empty,null,length,tail,init,append
                        ,inits,tails,reverse,transpose
                        ,concat,take,drop,splitAt,intercalate
-                       ,sort,isPrefixOf,isSuffixOf,isInfixOf,isSubstringOf
-                       ,findSubstring,findSubstrings,copy,group
+                       ,sort,isPrefixOf,isSuffixOf,isInfixOf
+                       ,findSubstring,findSubstrings,breakSubstring,copy,group
 
                        ,getLine, getContents, putStr, putStrLn, interact
                        ,hGetContents, hGet, hPut, hPutStr, hPutStrLn
@@ -243,10 +246,6 @@ import Data.ByteString (empty,null,length,tail,init,append
 import Data.ByteString.Internal (ByteString(PS), c2w, w2c, isSpaceWord8
                                 ,inlinePerformIO)
 
-#if defined(__GLASGOW_HASKELL__)
-import Data.ByteString.Unsafe (unsafePackAddress) -- for the rule
-#endif
-
 import Data.Char    ( isSpace )
 import qualified Data.List as List (intersperse)
 
@@ -260,12 +259,20 @@ import Foreign
 
 #if defined(__GLASGOW_HASKELL__)
 import GHC.Base                 (Char(..),unpackCString#,ord#,int2Word#)
-import GHC.IOBase               (IO(..),stToIO)
+#if __GLASGOW_HASKELL__ >= 611
+import GHC.IO                   (stToIO)
+#else
+import GHC.IOBase               (stToIO)
+#endif
 import GHC.Prim                 (Addr#,writeWord8OffAddr#,plusAddr#)
 import GHC.Ptr                  (Ptr(..))
 import GHC.ST                   (ST(..))
 #endif
 
+#if __GLASGOW_HASKELL__ >= 608
+import Data.String
+#endif
+
 #define STRICT1(f) f a | a `seq` False = undefined
 #define STRICT2(f) f a b | a `seq` b `seq` False = undefined
 #define STRICT3(f) f a b c | a `seq` b `seq` c `seq` False = undefined
@@ -278,6 +285,12 @@ singleton :: Char -> ByteString
 singleton = B.singleton . c2w
 {-# INLINE singleton #-}
 
+#if __GLASGOW_HASKELL__ >= 608
+instance IsString ByteString where
+    fromString = pack
+    {-# INLINE fromString #-}
+#endif
+
 -- | /O(n)/ Convert a 'String' into a 'ByteString'
 --
 -- For applications with large numbers of string literals, pack can be a
@@ -303,8 +316,8 @@ pack str = B.unsafeCreate (P.length str) $ \(Ptr p) -> stToIO (go p str)
 {-# INLINE [1] pack #-}
 
 {-# RULES
-    "FPS pack/packAddress" forall s .
-       pack (unpackCString# s) = inlinePerformIO (B.unsafePackAddress s)
+"ByteString pack/packAddress" forall s .
+   pack (unpackCString# s) = inlinePerformIO (B.unsafePackAddress s)
  #-}
 
 #endif
@@ -356,10 +369,6 @@ intersperse :: Char -> ByteString -> ByteString
 intersperse = B.intersperse . c2w
 {-# INLINE intersperse #-}
 
-join :: ByteString -> [ByteString] -> ByteString
-join = intercalate
-{-# DEPRECATED join "use intercalate" #-}
-
 -- | 'foldl', applied to a binary operator, a starting value (typically
 -- the left-identity of the operator), and a ByteString, reduces the
 -- ByteString using the binary operator, from left to right.
@@ -433,11 +442,6 @@ minimum :: ByteString -> Char
 minimum = w2c . B.minimum
 {-# INLINE minimum #-}
 
--- | /O(n)/ map Char functions, provided with the index at each position
-mapIndexed :: (Int -> Char -> Char) -> ByteString -> ByteString
-mapIndexed f = B.mapIndexed (\i c -> c2w (f i (w2c c)))
-{-# INLINE mapIndexed #-}
-
 -- | The 'mapAccumL' function behaves like a combination of 'map' and
 -- 'foldl'; it applies a function to each element of a ByteString,
 -- passing an accumulating parameter from left to right, and returning a
@@ -528,6 +532,11 @@ dropWhile f = B.dropWhile (f . w2c)
 {-# INLINE [1] dropWhile #-}
 #endif
 
+{-# RULES
+"ByteString specialise dropWhile isSpace -> dropSpace"
+    dropWhile isSpace = dropSpace
+  #-}
+
 -- | 'break' @p@ is equivalent to @'span' ('not' . p)@.
 break :: (Char -> Bool) -> ByteString -> (ByteString, ByteString)
 break f = B.break (f . w2c)
@@ -535,6 +544,30 @@ break f = B.break (f . w2c)
 {-# INLINE [1] break #-}
 #endif
 
+#if __GLASGOW_HASKELL__ >= 606
+-- This RULE LHS is not allowed by ghc-6.4
+{-# RULES
+"ByteString specialise break (x==)" forall x.
+    break ((==) x) = breakChar x
+"ByteString specialise break (==x)" forall x.
+    break (==x) = breakChar x
+  #-}
+#endif
+
+-- INTERNAL:
+
+-- | 'breakChar' breaks its ByteString argument at the first occurence
+-- of the specified char. It is more efficient than 'break' as it is
+-- implemented with @memchr(3)@. I.e.
+-- 
+-- > break (=='c') "abcd" == breakChar 'c' "abcd"
+--
+breakChar :: Char -> ByteString -> (ByteString, ByteString)
+breakChar c p = case elemIndex c p of
+    Nothing -> (p,empty)
+    Just n  -> (B.unsafeTake n p, B.unsafeDrop n p)
+{-# INLINE breakChar #-}
+
 -- | 'span' @p xs@ breaks the ByteString into two segments. It is
 -- equivalent to @('takeWhile' p xs, 'dropWhile' p xs)@
 span :: (Char -> Bool) -> ByteString -> (ByteString, ByteString)
@@ -703,6 +736,28 @@ filter :: (Char -> Bool) -> ByteString -> ByteString
 filter f = B.filter (f . w2c)
 {-# INLINE filter #-}
 
+{-
+-- | /O(n)/ and /O(n\/c) space/ A first order equivalent of /filter .
+-- (==)/, for the common case of filtering a single Char. It is more
+-- efficient to use /filterChar/ in this case.
+--
+-- > filterChar == filter . (==)
+--
+-- filterChar is around 10x faster, and uses much less space, than its
+-- filter equivalent
+--
+filterChar :: Char -> ByteString -> ByteString
+filterChar c ps = replicate (count c ps) c
+{-# INLINE filterChar #-}
+
+{-# RULES
+"ByteString specialise filter (== x)" forall x.
+    filter ((==) x) = filterChar x
+"ByteString specialise filter (== x)" forall x.
+    filter (== x) = filterChar x
+  #-}
+-}
+
 -- | /O(n)/ The 'find' function takes a predicate and a ByteString,
 -- and returns the first element in matching the predicate, or 'Nothing'
 -- if there is no such element.
@@ -773,8 +828,8 @@ unsafeHead  = w2c . B.unsafeHead
 -- Things that depend on the encoding
 
 {-# RULES
-    "FPS specialise break -> breakSpace"
-        break isSpace = breakSpace
+"ByteString specialise break -> breakSpace"
+    break isSpace = breakSpace
   #-}
 
 -- | 'breakSpace' returns the pair of ByteStrings when the argument is
@@ -799,11 +854,6 @@ firstspace ptr n m
     | otherwise = do w <- peekByteOff ptr n
                      if (not . isSpaceWord8) w then firstspace ptr (n+1) m else return n
 
-{-# RULES
-    "FPS specialise dropWhile isSpace -> dropSpace"
-        dropWhile isSpace = dropSpace
-  #-}
-
 -- | 'dropSpace' efficiently returns the 'ByteString' argument with
 -- white space Chars removed from the front. It is more efficient than
 -- calling dropWhile for removing whitespace. I.e.
@@ -855,7 +905,6 @@ lines ps
              Nothing -> [ps]
              Just n  -> take n ps : lines (drop (n+1) ps)
     where search = elemIndex '\n'
-{-# INLINE lines #-}
 
 {-
 -- Just as fast, but more complex. Should be much faster, I thought.
@@ -969,6 +1018,9 @@ readInteger as
           combine2 b (n:m:ns) = let t = m*b + n in t `seq` (t : combine2 b ns)
           combine2 _ ns       = ns
 
+------------------------------------------------------------------------
+-- For non-binary text processing:
+
 -- | Read an entire file strictly into a 'ByteString'.  This is far more
 -- efficient than reading the characters into a 'String' and then using
 -- 'pack'.  It also may be more efficient than opening the file and
