@@ -1,6 +1,6 @@
-{-# OPTIONS_GHC -cpp -fglasgow-exts -fno-warn-orphans -fno-warn-incomplete-patterns #-}
-
--- #prune
+{-# LANGUAGE CPP #-}
+{-# OPTIONS_GHC -fno-warn-incomplete-patterns #-}
+{-# OPTIONS_HADDOCK prune #-}
 
 -- |
 -- Module      : Data.ByteString.Lazy
@@ -8,7 +8,7 @@
 --               (c) Duncan Coutts 2006
 -- License     : BSD-style
 --
--- Maintainer  : dons@cse.unsw.edu.au
+-- Maintainer  : dons@galois.com
 -- Stability   : experimental
 -- Portability : portable
 -- 
@@ -101,7 +101,6 @@ module Data.ByteString.Lazy (
         -- ** Accumulating maps
         mapAccumL,              -- :: (acc -> Word8 -> (acc, Word8)) -> acc -> ByteString -> (acc, ByteString)
         mapAccumR,              -- :: (acc -> Word8 -> (acc, Word8)) -> acc -> ByteString -> (acc, ByteString)
-        mapIndexed,             -- :: (Int64 -> Word8 -> Word8) -> ByteString -> ByteString
 
         -- ** Infinite ByteStrings
         repeat,                 -- :: Word8 -> ByteString
@@ -193,16 +192,8 @@ module Data.ByteString.Lazy (
         hPut,                   -- :: Handle -> ByteString -> IO ()
         hPutStr,                -- :: Handle -> ByteString -> IO ()
 
---      hGetN,                  -- :: Int -> Handle -> Int -> IO ByteString
---      hGetContentsN,          -- :: Int -> Handle -> IO ByteString
---      hGetNonBlockingN,       -- :: Int -> Handle -> IO ByteString
-
-        -- undocumented deprecated things:
-        join                    -- :: ByteString -> [ByteString] -> ByteString
-
   ) where
 
-import qualified Prelude
 import Prelude hiding
     (reverse,head,tail,last,init,null,length,map,lines,foldl,foldr,unlines
     ,concat,any,take,drop,splitAt,takeWhile,dropWhile,span,break,elem,filter,maximum
@@ -211,11 +202,11 @@ import Prelude hiding
     ,getContents,getLine,putStr,putStrLn ,zip,zipWith,unzip,notElem)
 
 import qualified Data.List              as L  -- L for list/lazy
+import qualified Data.ByteString        as P  (ByteString) -- type name only
 import qualified Data.ByteString        as S  -- S for strict (hmm...)
 import qualified Data.ByteString.Internal as S
 import qualified Data.ByteString.Unsafe as S
 import Data.ByteString.Lazy.Internal
-import qualified Data.ByteString.Fusion as F
 
 import Data.Monoid              (Monoid(..))
 
@@ -223,6 +214,7 @@ import Data.Word                (Word8)
 import Data.Int                 (Int64)
 import System.IO                (Handle,stdin,stdout,openBinaryFile,IOMode(..)
                                 ,hClose,hWaitForInput,hIsEOF)
+import System.IO.Error          (mkIOError, illegalOperationErrorType)
 import System.IO.Unsafe
 #ifndef __NHC__
 import Control.Exception        (bracket)
@@ -290,15 +282,17 @@ cmp (Chunk a as) (Chunk b bs) =
 -- | /O(1)/ The empty 'ByteString'
 empty :: ByteString
 empty = Empty
+{-# INLINE empty #-}
 
 -- | /O(1)/ Convert a 'Word8' into a 'ByteString'
 singleton :: Word8 -> ByteString
 singleton w = Chunk (S.singleton w) Empty
+{-# INLINE singleton #-}
 
 -- | /O(n)/ Convert a '[Word8]' into a 'ByteString'. 
 pack :: [Word8] -> ByteString
 pack ws = L.foldr (Chunk . S.pack) Empty (chunks defaultChunkSize ws)
-  where 
+  where
     chunks :: Int -> [a] -> [[a]]
     chunks _    [] = []
     chunks size xs = case L.splitAt size xs of
@@ -310,11 +304,11 @@ unpack cs = L.concatMap S.unpack (toChunks cs)
 --TODO: we can do better here by integrating the concat with the unpack
 
 -- | /O(c)/ Convert a list of strict 'ByteString' into a lazy 'ByteString'
-fromChunks :: [S.ByteString] -> ByteString
+fromChunks :: [P.ByteString] -> ByteString
 fromChunks cs = L.foldr chunk Empty cs
 
 -- | /O(n)/ Convert a lazy 'ByteString' into a list of strict 'ByteString'
-toChunks :: ByteString -> [S.ByteString]
+toChunks :: ByteString -> [P.ByteString]
 toChunks cs = foldrChunks (:) [] cs
 
 ------------------------------------------------------------------------
@@ -341,15 +335,18 @@ unpackWith k (LPS ss) = L.concatMap (S.unpackWith k) ss
 null :: ByteString -> Bool
 null Empty = True
 null _     = False
+{-# INLINE null #-}
 
 -- | /O(n\/c)/ 'length' returns the length of a ByteString as an 'Int64'
 length :: ByteString -> Int64
 length cs = foldlChunks (\n c -> n + fromIntegral (S.length c)) 0 cs
+{-# INLINE length #-}
 
 -- | /O(1)/ 'cons' is analogous to '(:)' for lists.
 --
 cons :: Word8 -> ByteString -> ByteString
 cons c cs = Chunk (S.singleton c) cs
+{-# INLINE cons #-}
 
 -- | /O(1)/ Unlike 'cons', 'cons\'' is
 -- strict in the ByteString that we are consing onto. More precisely, it forces
@@ -367,10 +364,12 @@ cons c cs = Chunk (S.singleton c) cs
 cons' :: Word8 -> ByteString -> ByteString
 cons' w (Chunk c cs) | S.length c < 16 = Chunk (S.cons w c) cs
 cons' w cs                             = Chunk (S.singleton w) cs
+{-# INLINE cons' #-}
 
 -- | /O(n\/c)/ Append a byte to the end of a 'ByteString'
 snoc :: ByteString -> Word8 -> ByteString
 snoc cs w = foldrChunks Chunk (singleton w) cs
+{-# INLINE snoc #-}
 
 -- | /O(1)/ Extract the first element of a ByteString, which must be non-empty.
 head :: ByteString -> Word8
@@ -394,6 +393,7 @@ tail Empty          = errorEmptyList "tail"
 tail (Chunk c cs)
   | S.length c == 1 = cs
   | otherwise       = Chunk (S.unsafeTail c) cs
+{-# INLINE tail #-}
 
 -- | /O(n\/c)/ Extract the last element of a ByteString, which must be finite
 -- and non-empty.
@@ -402,6 +402,7 @@ last Empty          = errorEmptyList "last"
 last (Chunk c0 cs0) = go c0 cs0
   where go c Empty        = S.last c
         go _ (Chunk c cs) = go c cs
+-- XXX Don't inline this. Something breaks with 6.8.2 (haven't investigated yet)
 
 -- | /O(n\/c)/ Return all the elements of a 'ByteString' except the last one.
 init :: ByteString -> ByteString
@@ -414,6 +415,7 @@ init (Chunk c0 cs0) = go c0 cs0
 -- | /O(n\/c)/ Append two ByteStrings
 append :: ByteString -> ByteString -> ByteString
 append xs ys = foldrChunks Chunk ys xs
+{-# INLINE append #-}
 
 -- ---------------------------------------------------------------------
 -- Transformations
@@ -421,7 +423,13 @@ append xs ys = foldrChunks Chunk ys xs
 -- | /O(n)/ 'map' @f xs@ is the ByteString obtained by applying @f@ to each
 -- element of @xs@.
 map :: (Word8 -> Word8) -> ByteString -> ByteString
-map f = F.loopArr . F.loopL (F.mapEFL f) F.NoAcc
+map f s = go s
+    where
+        go Empty        = Empty
+        go (Chunk x xs) = Chunk y ys
+            where
+                y  = S.map f x
+                ys = go xs
 {-# INLINE map #-}
 
 -- | /O(n)/ 'reverse' @xs@ returns the elements of @xs@ in reverse order.
@@ -438,7 +446,7 @@ intersperse :: Word8 -> ByteString -> ByteString
 intersperse _ Empty        = Empty
 intersperse w (Chunk c cs) = Chunk (S.intersperse w c)
                                    (foldrChunks (Chunk . intersperse') Empty cs)
-  where intersperse' :: S.ByteString -> S.ByteString
+  where intersperse' :: P.ByteString -> P.ByteString
         intersperse' (S.PS fp o l) =
           S.unsafeCreate (2*l) $ \p' -> withForeignPtr fp $ \p -> do
             poke p' w
@@ -458,12 +466,17 @@ transpose css = L.map (\ss -> Chunk (S.pack ss) Empty)
 -- the left-identity of the operator), and a ByteString, reduces the
 -- ByteString using the binary operator, from left to right.
 foldl :: (a -> Word8 -> a) -> a -> ByteString -> a
-foldl f z = F.loopAcc . F.loopL (F.foldEFL f) z
+foldl f z = go z
+  where go a Empty        = a
+        go a (Chunk c cs) = go (S.foldl f a c) cs
 {-# INLINE foldl #-}
 
 -- | 'foldl\'' is like 'foldl', but strict in the accumulator.
 foldl' :: (a -> Word8 -> a) -> a -> ByteString -> a
-foldl' f z = F.loopAcc . F.loopL (F.foldEFL' f) z
+foldl' f z = go z
+  where go a _ | a `seq` False = undefined
+        go a Empty        = a
+        go a (Chunk c cs) = go (S.foldl f a c) cs
 {-# INLINE foldl' #-}
 
 -- | 'foldr', applied to a binary operator, a starting value
@@ -510,12 +523,12 @@ concatMap :: (Word8 -> ByteString) -> ByteString -> ByteString
 concatMap _ Empty        = Empty
 concatMap f (Chunk c0 cs0) = to c0 cs0
   where
-    go :: ByteString -> S.ByteString -> ByteString -> ByteString
+    go :: ByteString -> P.ByteString -> ByteString -> ByteString
     go Empty        c' cs' = to c' cs'
     go (Chunk c cs) c' cs' = Chunk c (go cs c' cs')
 
-    to :: S.ByteString -> ByteString -> ByteString
-    to c cs | S.null c  = case cs of 
+    to :: P.ByteString -> ByteString -> ByteString
+    to c cs | S.null c  = case cs of
         Empty          -> Empty
         (Chunk c' cs') -> to c' cs'
             | otherwise = go (f (S.unsafeHead c)) (S.unsafeTail c) cs
@@ -524,12 +537,14 @@ concatMap f (Chunk c0 cs0) = to c0 cs0
 -- any element of the 'ByteString' satisfies the predicate.
 any :: (Word8 -> Bool) -> ByteString -> Bool
 any f cs = foldrChunks (\c rest -> S.any f c || rest) False cs
+{-# INLINE any #-}
 -- todo fuse
 
 -- | /O(n)/ Applied to a predicate and a 'ByteString', 'all' determines
 -- if all elements of the 'ByteString' satisfy the predicate.
 all :: (Word8 -> Bool) -> ByteString -> Bool
 all f cs = foldrChunks (\c rest -> S.all f c && rest) True cs
+{-# INLINE all #-}
 -- todo fuse
 
 -- | /O(n)/ 'maximum' returns the maximum value from a 'ByteString'
@@ -570,10 +585,6 @@ mapAccumR f s0 cs0 = go s0 cs0
         where (s'', c') = S.mapAccumR f s' c
               (s', cs') = go s cs
 
--- | /O(n)/ map Word8 functions, provided with the index at each position
-mapIndexed :: (Int -> Word8 -> Word8) -> ByteString -> ByteString
-mapIndexed f = F.loopArr . F.loopL (F.mapIndexEFL f) 0
-
 -- ---------------------------------------------------------------------
 -- Building ByteStrings
 
@@ -586,7 +597,9 @@ mapIndexed f = F.loopArr . F.loopL (F.mapIndexEFL f) 0
 --
 -- > last (scanl f z xs) == foldl f z xs.
 scanl :: (Word8 -> Word8 -> Word8) -> Word8 -> ByteString -> ByteString
-scanl f z ps = F.loopArr . F.loopL (F.scanEFL f) z $ (ps `snoc` 0)
+scanl f z = snd . foldl k (z,singleton z)
+ where
+    k (c,acc) a = let n = f c a in (n, acc `snoc` n)
 {-# INLINE scanl #-}
 
 -- ---------------------------------------------------------------------
@@ -621,7 +634,6 @@ replicate n w
     (q, r) = quotRem n (fromIntegral smallChunkSize)
     nChunks 0 = Empty
     nChunks m = Chunk c (nChunks (m-1))
-    
 
 -- | 'cycle' ties a finite ByteString into a circular one, or equivalently,
 -- the infinite repetition of the original ByteString.
@@ -776,7 +788,7 @@ splitWith :: (Word8 -> Bool) -> ByteString -> [ByteString]
 splitWith _ Empty          = []
 splitWith p (Chunk c0 cs0) = comb [] (S.splitWith p c0) cs0
 
-  where comb :: [S.ByteString] -> [S.ByteString] -> ByteString -> [ByteString]
+  where comb :: [P.ByteString] -> [P.ByteString] -> ByteString -> [ByteString]
         comb acc (s:[]) Empty        = revChunks (s:acc) : []
         comb acc (s:[]) (Chunk c cs) = comb (s:acc) (S.splitWith p c) cs
         comb acc (s:ss) cs           = revChunks (s:acc) : comb [] ss cs
@@ -803,7 +815,7 @@ split :: Word8 -> ByteString -> [ByteString]
 split _ Empty     = []
 split w (Chunk c0 cs0) = comb [] (S.split w c0) cs0
 
-  where comb :: [S.ByteString] -> [S.ByteString] -> ByteString -> [ByteString]
+  where comb :: [P.ByteString] -> [P.ByteString] -> ByteString -> [ByteString]
         comb acc (s:[]) Empty        = revChunks (s:acc) : []
         comb acc (s:[]) (Chunk c cs) = comb (s:acc) (S.split w c) cs
         comb acc (s:ss) cs           = revChunks (s:acc) : comb [] ss cs
@@ -832,7 +844,7 @@ group :: ByteString -> [ByteString]
 group Empty          = []
 group (Chunk c0 cs0) = group' [] (S.group c0) cs0
   where 
-    group' :: [S.ByteString] -> [S.ByteString] -> ByteString -> [ByteString]
+    group' :: [P.ByteString] -> [P.ByteString] -> ByteString -> [ByteString]
     group' acc@(s':_) ss@(s:_) cs
       | S.unsafeHead s'
      /= S.unsafeHead s             = revNonEmptyChunks    acc  : group' [] ss cs
@@ -857,7 +869,7 @@ groupBy :: (Word8 -> Word8 -> Bool) -> ByteString -> [ByteString]
 groupBy _ Empty          = []
 groupBy k (Chunk c0 cs0) = groupBy' [] 0 (S.groupBy k c0) cs0
   where
-    groupBy' :: [S.ByteString] -> Word8 -> [S.ByteString] -> ByteString -> [ByteString]
+    groupBy' :: [P.ByteString] -> Word8 -> [P.ByteString] -> ByteString -> [ByteString]
     groupBy' acc@(_:_) c ss@(s:_) cs
       | not (c `k` S.unsafeHead s)     = revNonEmptyChunks acc : groupBy' [] 0 ss cs
     groupBy' acc _ (s:[]) Empty        = revNonEmptyChunks (s : acc) : []
@@ -883,10 +895,6 @@ groupBy k xs
 intercalate :: ByteString -> [ByteString] -> ByteString
 intercalate s = concat . (L.intersperse s)
 
-join :: ByteString -> [ByteString] -> ByteString
-join = intercalate
-{-# DEPRECATED join "use intercalate" #-}
-
 -- ---------------------------------------------------------------------
 -- Indexing ByteStrings
 
@@ -998,7 +1006,10 @@ notElem w cs = not (elem w cs)
 -- returns a ByteString containing those characters that satisfy the
 -- predicate.
 filter :: (Word8 -> Bool) -> ByteString -> ByteString
-filter p = F.loopArr . F.loopL (F.filterEFL p) F.NoAcc
+filter p s = go s
+    where
+        go Empty        = Empty
+        go (Chunk x xs) = chunk (S.filter p x) (go xs)
 {-# INLINE filter #-}
 
 {-
@@ -1012,8 +1023,18 @@ filter p = F.loopArr . F.loopL (F.filterEFL p) F.NoAcc
 -- filter equivalent
 filterByte :: Word8 -> ByteString -> ByteString
 filterByte w ps = replicate (count w ps) w
--- filterByte w (LPS xs) = LPS (filterMap (P.filterByte w) xs)
+{-# INLINE filterByte #-}
 
+{-# RULES
+"ByteString specialise filter (== x)" forall x.
+  filter ((==) x) = filterByte x
+
+"ByteString specialise filter (== x)" forall x.
+ filter (== x) = filterByte x
+  #-}
+-}
+
+{-
 -- | /O(n)/ A first order equivalent of /filter . (\/=)/, for the common
 -- case of filtering a single byte out of a list. It is more efficient
 -- to use /filterNotByte/ in this case.
@@ -1135,13 +1156,20 @@ copy cs = foldrChunks (Chunk . S.copy) Empty cs
 
 -- ---------------------------------------------------------------------
 -- Lazy ByteString IO
+--
+-- Rule for when to close: is it expected to read the whole file?
+-- If so, close when done. 
+--
 
 -- | Read entire handle contents /lazily/ into a 'ByteString'. Chunks
 -- are read on demand, in at most @k@-sized chunks. It does not block
 -- waiting for a whole @k@-sized chunk, so if less than @k@ bytes are
 -- available then they will be returned immediately as a smaller chunk.
+--
+-- The handle is closed on EOF.
+--
 hGetContentsN :: Int -> Handle -> IO ByteString
-hGetContentsN k h = lazyRead
+hGetContentsN k h = lazyRead -- TODO close on exceptions
   where
     lazyRead = unsafeInterleaveIO loop
 
@@ -1152,7 +1180,7 @@ hGetContentsN k h = lazyRead
         -- 0 or EAGAIN
         if S.null c
           then do eof <- hIsEOF h
-                  if eof then return Empty
+                  if eof then hClose h >> return Empty
                          else hWaitForInput h (-1)
                            >> loop
           else do cs <- lazyRead
@@ -1160,9 +1188,9 @@ hGetContentsN k h = lazyRead
 
 -- | Read @n@ bytes into a 'ByteString', directly from the
 -- specified 'Handle', in chunks of size @k@.
+--
 hGetN :: Int -> Handle -> Int -> IO ByteString
-hGetN _ _ 0 = return empty
-hGetN k h n = readChunks n
+hGetN k h n | n > 0 = readChunks n
   where
     STRICT1(readChunks)
     readChunks i = do
@@ -1172,13 +1200,16 @@ hGetN k h n = readChunks n
             m -> do cs <- readChunks (i - m)
                     return (Chunk c cs)
 
+hGetN _ _ 0 = return Empty
+hGetN _ h n = illegalBufferSize h "hGet" n
+
 -- | hGetNonBlockingN is similar to 'hGetContentsN', except that it will never block
 -- waiting for data to become available, instead it returns only whatever data
 -- is available. Chunks are read on demand, in @k@-sized chunks.
+--
 hGetNonBlockingN :: Int -> Handle -> Int -> IO ByteString
 #if defined(__GLASGOW_HASKELL__)
-hGetNonBlockingN _ _ 0 = return empty
-hGetNonBlockingN k h n = readChunks n
+hGetNonBlockingN k h n | n > 0= readChunks n
   where
     STRICT1(readChunks)
     readChunks i = do
@@ -1187,16 +1218,30 @@ hGetNonBlockingN k h n = readChunks n
             0 -> return Empty
             m -> do cs <- readChunks (i - m)
                     return (Chunk c cs)
+
+hGetNonBlockingN _ _ 0 = return Empty
+hGetNonBlockingN _ h n = illegalBufferSize h "hGetNonBlocking" n
 #else
 hGetNonBlockingN = hGetN
 #endif
 
+illegalBufferSize :: Handle -> String -> Int -> IO a
+illegalBufferSize handle fn sz =
+    ioError (mkIOError illegalOperationErrorType msg (Just handle) Nothing)
+    --TODO: System.IO uses InvalidArgument here, but it's not exported :-(
+    where
+      msg = fn ++ ": illegal ByteString size " ++ showsPrec 9 sz []
+
 -- | Read entire handle contents /lazily/ into a 'ByteString'. Chunks
 -- are read on demand, using the default chunk size.
+--
+-- Once EOF is encountered, the Handle is closed.
+--
 hGetContents :: Handle -> IO ByteString
 hGetContents = hGetContentsN defaultChunkSize
 
 -- | Read @n@ bytes into a 'ByteString', directly from the specified 'Handle'.
+--
 hGet :: Handle -> Int -> IO ByteString
 hGet = hGetN defaultChunkSize
 
@@ -1211,28 +1256,35 @@ hGetNonBlocking = hGet
 #endif
 
 -- | Read an entire file /lazily/ into a 'ByteString'.
+-- The Handle will be held open until EOF is encountered.
+--
 readFile :: FilePath -> IO ByteString
 readFile f = openBinaryFile f ReadMode >>= hGetContents
 
 -- | Write a 'ByteString' to a file.
+--
 writeFile :: FilePath -> ByteString -> IO ()
 writeFile f txt = bracket (openBinaryFile f WriteMode) hClose
     (\hdl -> hPut hdl txt)
 
 -- | Append a 'ByteString' to a file.
+--
 appendFile :: FilePath -> ByteString -> IO ()
 appendFile f txt = bracket (openBinaryFile f AppendMode) hClose
     (\hdl -> hPut hdl txt)
 
 -- | getContents. Equivalent to hGetContents stdin. Will read /lazily/
+--
 getContents :: IO ByteString
 getContents = hGetContents stdin
 
 -- | Outputs a 'ByteString' to the specified 'Handle'.
+--
 hPut :: Handle -> ByteString -> IO ()
 hPut h cs = foldrChunks (\c rest -> S.hPut h c >> rest) (return ()) cs
 
 -- | A synonym for @hPut@, for compatibility
+--
 hPutStr :: Handle -> ByteString -> IO ()
 hPutStr = hPut
 
@@ -1241,13 +1293,15 @@ putStr :: ByteString -> IO ()
 putStr = hPut stdout
 
 -- | Write a ByteString to stdout, appending a newline byte
+--
 putStrLn :: ByteString -> IO ()
 putStrLn ps = hPut stdout ps >> hPut stdout (singleton 0x0a)
 
 -- | The interact function takes a function of type @ByteString -> ByteString@
 -- as its argument. The entire input from the standard input device is passed
 -- to this function as its argument, and the resulting string is output on the
--- standard output device. It's great for writing one line programs!
+-- standard output device.
+--
 interact :: (ByteString -> ByteString) -> IO ()
 interact transformer = putStr . transformer =<< getContents
 
@@ -1264,16 +1318,16 @@ moduleError fun msg = error ("Data.ByteString.Lazy." ++ fun ++ ':':' ':msg)
 
 
 -- reverse a list of non-empty chunks into a lazy ByteString
-revNonEmptyChunks :: [S.ByteString] -> ByteString
+revNonEmptyChunks :: [P.ByteString] -> ByteString
 revNonEmptyChunks cs = L.foldl' (flip Chunk) Empty cs
 
 -- reverse a list of possibly-empty chunks into a lazy ByteString
-revChunks :: [S.ByteString] -> ByteString
+revChunks :: [P.ByteString] -> ByteString
 revChunks cs = L.foldl' (flip chunk) Empty cs
 
 -- | 'findIndexOrEnd' is a variant of findIndex, that returns the length
 -- of the string if no element is found, rather than Nothing.
-findIndexOrEnd :: (Word8 -> Bool) -> S.ByteString -> Int
+findIndexOrEnd :: (Word8 -> Bool) -> P.ByteString -> Int
 findIndexOrEnd k (S.PS x s l) = S.inlinePerformIO $ withForeignPtr x $ \f -> go (f `plusPtr` s) 0
   where
     STRICT2(go)
