@@ -1,6 +1,5 @@
-{-# OPTIONS_GHC -cpp -fno-warn-orphans #-}
-
--- #prune
+{-# LANGUAGE CPP #-}
+{-# OPTIONS_HADDOCK prune #-}
 
 -- |
 -- Module      : Data.ByteString.Lazy.Char8
@@ -80,7 +79,7 @@ module Data.ByteString.Lazy.Char8 (
 
         -- ** Accumulating maps
         mapAccumL,              -- :: (acc -> Char -> (acc, Char)) -> acc -> ByteString -> (acc, ByteString)
-        mapIndexed,             -- :: (Int64 -> Char -> Char) -> ByteString -> ByteString
+        mapAccumR,              -- :: (acc -> Char -> (acc, Char)) -> acc -> ByteString -> (acc, ByteString)
 
         -- ** Infinite ByteStrings
         repeat,                 -- :: Char -> ByteString
@@ -174,18 +173,11 @@ module Data.ByteString.Lazy.Char8 (
         hGetNonBlocking,        -- :: Handle -> Int64 -> IO ByteString
         hPut,                   -- :: Handle -> ByteString -> IO ()
 
---      hGetN,                  -- :: Int -> Handle -> Int64 -> IO ByteString
---      hGetContentsN,          -- :: Int -> Handle -> IO ByteString
---      hGetNonBlockingN,       -- :: Int -> Handle -> IO ByteString
-
-        -- undocumented deprecated things:
-        join                    -- :: ByteString -> [ByteString] -> ByteString
-
   ) where
 
 -- Functions transparently exported
 import Data.ByteString.Lazy 
-        (ByteString, fromChunks, toChunks
+        (fromChunks, toChunks
         ,empty,null,length,tail,init,append,reverse,transpose,cycle
         ,concat,take,drop,splitAt,intercalate,isPrefixOf,group,inits,tails,copy
         ,hGetContents, hGet, hPut, getContents
@@ -194,8 +186,8 @@ import Data.ByteString.Lazy
 
 -- Functions we need to wrap.
 import qualified Data.ByteString.Lazy as L
+import qualified Data.ByteString as S (ByteString) -- typename only
 import qualified Data.ByteString as B
-import qualified Data.ByteString.Internal as B
 import qualified Data.ByteString.Unsafe as B
 import Data.ByteString.Lazy.Internal
 
@@ -218,11 +210,16 @@ import Control.Exception    (bracket)
 import IO                   (bracket)
 #endif
 
+#if __GLASGOW_HASKELL__ >= 608
+import Data.String
+#endif
+
 #define STRICT1(f) f a | a `seq` False = undefined
 #define STRICT2(f) f a b | a `seq` b `seq` False = undefined
 #define STRICT3(f) f a b c | a `seq` b `seq` c `seq` False = undefined
 #define STRICT4(f) f a b c d | a `seq` b `seq` c `seq` d `seq` False = undefined
 #define STRICT5(f) f a b c d e | a `seq` b `seq` c `seq` d `seq` e `seq` False = undefined
+#define STRICT5_(f) f a b c d _ | a `seq` b `seq` c `seq` d `seq` False = undefined
 
 ------------------------------------------------------------------------
 
@@ -231,6 +228,12 @@ singleton :: Char -> ByteString
 singleton = L.singleton . c2w
 {-# INLINE singleton #-}
 
+#if __GLASGOW_HASKELL__ >= 608
+instance IsString ByteString where
+    fromString = pack
+    {-# INLINE fromString #-}
+#endif
+
 -- | /O(n)/ Convert a 'String' into a 'ByteString'. 
 pack :: [Char] -> ByteString
 pack = L.pack. List.map c2w
@@ -298,10 +301,6 @@ intersperse :: Char -> ByteString -> ByteString
 intersperse = L.intersperse . c2w
 {-# INLINE intersperse #-}
 
-join :: ByteString -> [ByteString] -> ByteString
-join = intercalate
-{-# DEPRECATED join "use intercalate" #-}
-
 -- | 'foldl', applied to a binary operator, a starting value (typically
 -- the left-identity of the operator), and a ByteString, reduces the
 -- ByteString using the binary operator, from left to right.
@@ -385,9 +384,12 @@ scanl f z = L.scanl (\a b -> c2w (f (w2c a) (w2c b))) (c2w z)
 mapAccumL :: (acc -> Char -> (acc, Char)) -> acc -> ByteString -> (acc, ByteString)
 mapAccumL f = L.mapAccumL (\a w -> case f a (w2c w) of (a',c) -> (a', c2w c))
 
--- | /O(n)/ map Char functions, provided with the index at each position
-mapIndexed :: (Int -> Char -> Char) -> ByteString -> ByteString
-mapIndexed f = L.mapIndexed (\i w -> c2w (f i (w2c w)))
+-- | The 'mapAccumR' function behaves like a combination of 'map' and
+-- 'foldr'; it applies a function to each element of a ByteString,
+-- passing an accumulating parameter from right to left, and returning a
+-- final value of this accumulator together with the new ByteString.
+mapAccumR :: (acc -> Char -> (acc, Char)) -> acc -> ByteString -> (acc, ByteString)
+mapAccumR f = L.mapAccumR (\acc w -> case f acc (w2c w) of (acc', c) -> (acc', c2w c))
 
 ------------------------------------------------------------------------
 -- Generating and unfolding ByteStrings
@@ -565,6 +567,31 @@ filter :: (Char -> Bool) -> ByteString -> ByteString
 filter f = L.filter (f . w2c)
 {-# INLINE filter #-}
 
+{-
+-- | /O(n)/ and /O(n\/c) space/ A first order equivalent of /filter .
+-- (==)/, for the common case of filtering a single Char. It is more
+-- efficient to use /filterChar/ in this case.
+--
+-- > filterChar == filter . (==)
+--
+-- filterChar is around 10x faster, and uses much less space, than its
+-- filter equivalent
+--
+filterChar :: Char -> ByteString -> ByteString
+filterChar c ps = replicate (count c ps) c
+{-# INLINE filterChar #-}
+
+{-# RULES
+  "ByteString specialise filter (== x)" forall x.
+      filter ((==) x) = filterChar x
+  #-}
+
+{-# RULES
+  "ByteString specialise filter (== x)" forall x.
+     filter (== x) = filterChar x
+  #-}
+-}
+
 -- | /O(n)/ The 'find' function takes a predicate and a ByteString,
 -- and returns the first element in matching the predicate, or 'Nothing'
 -- if there is no such element.
@@ -620,6 +647,9 @@ zipWith f = L.zipWith ((. w2c) . f . w2c)
 -- | 'lines' breaks a ByteString up into a list of ByteStrings at
 -- newline Chars. The resulting strings do not contain newlines.
 --
+-- As of bytestring 0.9.0.3, this function is stricter than its 
+-- list cousin.
+--
 lines :: ByteString -> [ByteString]
 lines Empty          = []
 lines (Chunk c0 cs0) = loop0 c0 cs0
@@ -635,8 +665,7 @@ lines (Chunk c0 cs0) = loop0 c0 cs0
 
     -- the common special case where we have no existing chunks of
     -- the current line
-    loop0 :: B.ByteString -> ByteString -> [ByteString]
-    STRICT2(loop0)
+    loop0 :: S.ByteString -> ByteString -> [ByteString]
     loop0 c cs =
         case B.elemIndex (c2w '\n') c of
             Nothing -> case cs of
@@ -653,8 +682,7 @@ lines (Chunk c0 cs0) = loop0 c0 cs0
 
     -- the general case when we are building a list of chunks that are
     -- part of the same line
-    loop :: B.ByteString -> [B.ByteString] -> ByteString -> [ByteString]
-    STRICT3(loop)
+    loop :: S.ByteString -> [S.ByteString] -> ByteString -> [ByteString]
     loop c line cs =
         case B.elemIndex (c2w '\n') c of
             Nothing ->
@@ -668,6 +696,30 @@ lines (Chunk c0 cs0) = loop0 c0 cs0
                 let c' = revChunks (B.unsafeTake n c : line)
                  in c' `seq` (c' : loop0 (B.unsafeDrop (n+1) c) cs)
 
+{-
+
+This function is too strict!  Consider,
+
+> prop_lazy =
+    (L.unpack . head . lazylines $ L.append (L.pack "a\nb\n") (error "failed"))
+  ==
+    "a"
+
+fails.  Here's a properly lazy version of 'lines' for lazy bytestrings
+
+    lazylines           :: L.ByteString -> [L.ByteString]
+    lazylines s
+        | L.null s  = []
+        | otherwise =
+            let (l,s') = L.break ((==) '\n') s
+            in l : if L.null s' then []
+                                else lazylines (L.tail s')
+
+we need a similarly lazy, but efficient version.
+
+-}
+
+
 -- | 'unlines' is an inverse operation to 'lines'.  It joins lines,
 -- after appending a terminating newline to each.
 unlines :: [ByteString] -> ByteString
@@ -693,16 +745,25 @@ unwords = intercalate (singleton ' ')
 -- there is no integer at the beginning of the string, it returns
 -- Nothing, otherwise it just returns the int read, and the rest of the
 -- string.
-readInt :: ByteString -> Maybe (Int, ByteString)
-readInt Empty        = Nothing
-readInt (Chunk x xs) =
-        case w2c (B.unsafeHead x) of
-            '-' -> loop True  0 0 (B.unsafeTail x) xs
-            '+' -> loop False 0 0 (B.unsafeTail x) xs
-            _   -> loop False 0 0 x xs
 
-    where loop :: Bool -> Int -> Int -> B.ByteString -> ByteString -> Maybe (Int, ByteString)
-          STRICT5(loop)
+{-
+-- Faster:
+
+data MaybeS = NothingS
+            | JustS {-# UNPACK #-} !Int {-# UNPACK #-} !ByteString
+-}
+
+readInt :: ByteString -> Maybe (Int, ByteString)
+{-# INLINE readInt #-}
+readInt Empty        = Nothing
+readInt (Chunk x xs) = case w2c (B.unsafeHead x) of
+    '-' -> loop True  0 0 (B.unsafeTail x) xs
+    '+' -> loop False 0 0 (B.unsafeTail x) xs
+    _   -> loop False 0 0 x xs
+
+    where loop :: Bool -> Int -> Int
+                -> S.ByteString -> ByteString -> Maybe (Int, ByteString)
+          STRICT5_(loop)
           loop neg i n c cs
               | B.null c = case cs of
                              Empty          -> end  neg i n c  cs
@@ -715,11 +776,13 @@ readInt (Chunk x xs) =
                                           (B.unsafeTail c) cs
                       | otherwise -> end neg i n c cs
 
-          end _   0 _ _  _   = Nothing
-          end neg _ n c cs = let n' | neg       = negate n
-                                    | otherwise = n
-                                 c' = chunk c cs
-                              in n' `seq` c' `seq` Just $! (n', c')
+          {-# INLINE end #-}
+          end _   0 _ _  _ = Nothing
+          end neg _ n c cs = e `seq` e
+                where n' = if neg then negate n else n
+                      c' = chunk c cs
+                      e  = n' `seq` c' `seq` Just $! (n',c')
+         --                  in n' `seq` c' `seq` JustS n' c'
 
 
 -- | readInteger reads an Integer from the beginning of the ByteString.  If
@@ -745,8 +808,8 @@ readInteger (Chunk c0 cs0) =
                 | otherwise              -> Nothing
 
           loop :: Int -> Int -> [Integer]
-               -> B.ByteString -> ByteString -> (Integer, ByteString)
-          STRICT5(loop)
+               -> S.ByteString -> ByteString -> (Integer, ByteString)
+          STRICT5_(loop)
           loop d acc ns c cs
               | B.null c = case cs of
                              Empty          -> combine d acc ns c cs
@@ -795,5 +858,5 @@ appendFile f txt = bracket (openFile f AppendMode) hClose
 -- Internal utilities
 
 -- reverse a list of possibly-empty chunks into a lazy ByteString
-revChunks :: [B.ByteString] -> ByteString
+revChunks :: [S.ByteString] -> ByteString
 revChunks cs = List.foldl' (flip chunk) Empty cs
