@@ -18,10 +18,6 @@
 --
 -- @'takeDirectory' file '</>' \"interface\" '</>' ('takeFileName' file \`replaceExtension\` \"hi\"@)
 --
--- You want to display a filename to the user, as neatly as possible
---
--- @'makeRelativeToCurrentDirectory' file >>= putStrLn@
---
 -- The examples in code format descibed by each function are used to generate
 -- tests, and should give clear semantics for the functions.
 -----------------------------------------------------------------------------
@@ -46,12 +42,9 @@ module System.FilePath.MODULE_NAME
     takeExtension, replaceExtension, dropExtension, addExtension, hasExtension, (<.>),
     splitExtensions, dropExtensions, takeExtensions,
 
-    -- Note: leave this section to enable some of the tests to work
-#ifdef TESTING
     -- * Drive methods
     splitDrive, joinDrive,
-    takeDrive, replaceDrive, hasDrive, dropDrive, isDrive,
-#endif
+    takeDrive, hasDrive, dropDrive, isDrive,
 
     -- * Operations on a FilePath, as a list of directories
     splitFileName,
@@ -68,17 +61,21 @@ module System.FilePath.MODULE_NAME
 
     -- * File name manipulators
     normalise, equalFilePath,
-    makeRelativeToCurrentDirectory, makeRelative,
+    makeRelative,
     isRelative, isAbsolute,
     isValid, makeValid
+
+#ifdef TESTING
+    , isRelativeDrive
+#endif
+
     )
     where
 
-import Data.Maybe(isJust, fromJust)
 import Data.Char(toLower, toUpper)
+import Data.Maybe(isJust, fromJust)
 
 import System.Environment(getEnv)
-import System.Directory(getCurrentDirectory)
 
 
 infixr 7  <.>
@@ -162,16 +159,23 @@ isExtSeparator = (== extSeparator)
 
 -- | Take a string, split it on the 'searchPathSeparator' character.
 --
--- > Windows: splitSearchPath "File1;File2;File3" == ["File1","File2","File3"]
--- > Posix:   splitSearchPath "File1:File2:File3" == ["File1","File2","File3"]
+--   Follows the recommendations in
+--   <http://www.opengroup.org/onlinepubs/009695399/basedefs/xbd_chap08.html>
+--
+-- > Posix:   splitSearchPath "File1:File2:File3"  == ["File1","File2","File3"]
+-- > Posix:   splitSearchPath "File1::File2:File3" == ["File1",".","File2","File3"]
+-- > Windows: splitSearchPath "File1;File2;File3"  == ["File1","File2","File3"]
+-- > Windows: splitSearchPath "File1;;File2;File3" == ["File1","File2","File3"]
 splitSearchPath :: String -> [FilePath]
 splitSearchPath = f
     where
     f xs = case break isSearchPathSeparator xs of
-           ([],  [])   -> []
-           ([],  post) -> f (tail post)
-           (pre, [])   -> [pre]
-           (pre, post) -> pre : f (tail post)
+           (pre, []    ) -> g pre
+           (pre, _:post) -> g pre ++ f post
+
+    g "" = ["." | isPosix]
+    g x = [x]
+
 
 -- | Get a list of filepaths in the $PATH.
 getSearchPath :: IO [FilePath]
@@ -203,8 +207,8 @@ splitExtension x = case d of
 -- | Get the extension of a file, returns @\"\"@ for no extension, @.ext@ otherwise.
 --
 -- > takeExtension x == snd (splitExtension x)
--- > takeExtension (addExtension x "ext") == ".ext"
--- > takeExtension (replaceExtension x "ext") == ".ext"
+-- > Valid x => takeExtension (addExtension x "ext") == ".ext"
+-- > Valid x => takeExtension (replaceExtension x "ext") == ".ext"
 takeExtension :: FilePath -> String
 takeExtension = snd . splitExtension
 
@@ -235,7 +239,7 @@ dropExtension = fst . splitExtension
 -- > addExtension "file." ".bib" == "file..bib"
 -- > addExtension "file" ".bib" == "file.bib"
 -- > addExtension "/" "x" == "/.x"
--- > takeFileName (addExtension (addTrailingPathSeparator x) "ext") == ".ext"
+-- > Valid x => takeFileName (addExtension (addTrailingPathSeparator x) "ext") == ".ext"
 -- > Windows: addExtension "\\\\share" ".txt" == "\\\\share\\.txt"
 addExtension :: FilePath -> String -> FilePath
 addExtension file "" = file
@@ -269,6 +273,8 @@ dropExtensions :: FilePath -> FilePath
 dropExtensions = fst . splitExtensions
 
 -- | Get all extensions
+--
+-- > takeExtensions "file.tar.gz" == ".tar.gz"
 takeExtensions :: FilePath -> String
 takeExtensions = snd . splitExtensions
 
@@ -293,8 +299,9 @@ isLetter x = (x >= 'a' && x <= 'z') || (x >= 'A' && x <= 'Z')
 -- > Windows: splitDrive "\\\\shared\\test" == ("\\\\shared\\","test")
 -- > Windows: splitDrive "\\\\shared" == ("\\\\shared","")
 -- > Windows: splitDrive "\\\\?\\UNC\\shared\\file" == ("\\\\?\\UNC\\shared\\","file")
+-- > Windows: splitDrive "\\\\?\\UNCshared\\file" == ("\\\\?\\","UNCshared\\file")
 -- > Windows: splitDrive "\\\\?\\d:\\file" == ("\\\\?\\d:\\","file")
--- > Windows: splitDrive "/d" == ("/","d")
+-- > Windows: splitDrive "/d" == ("","/d")
 -- > Posix:   splitDrive "/test" == ("/","test")
 -- > Posix:   splitDrive "//test" == ("//","test")
 -- > Posix:   splitDrive "test/file" == ("","test/file")
@@ -311,10 +318,9 @@ splitDrive x | isJust y = fromJust y
 splitDrive x | isJust y = fromJust y
     where y = readDriveShare x
 
-splitDrive (x:xs) | isPathSeparator x = addSlash [x] xs
-
 splitDrive x = ("",x)
 
+addSlash :: FilePath -> FilePath -> (FilePath, FilePath)
 addSlash a xs = (a++c,d)
     where (c,d) = span isPathSeparator xs
 
@@ -330,23 +336,23 @@ readDriveUNC (s1:s2:'?':s3:xs) | all isPathSeparator [s1,s2,s3] =
         _ -> case readDriveLetter xs of
                  Just (a,b) -> Just (s1:s2:'?':s3:a,b)
                  Nothing -> Nothing
-readDriveUNC x = Nothing
+readDriveUNC _ = Nothing
 
--- c:\
+{- c:\ -}
 readDriveLetter :: String -> Maybe (FilePath, FilePath)
 readDriveLetter (x:':':y:xs) | isLetter x && isPathSeparator y = Just $ addSlash [x,':'] (y:xs)
 readDriveLetter (x:':':xs) | isLetter x = Just ([x,':'], xs)
-readDriveLetter x = Nothing
+readDriveLetter _ = Nothing
 
--- \\sharename\
+{- \\sharename\ -}
 readDriveShare :: String -> Maybe (FilePath, FilePath)
 readDriveShare (s1:s2:xs) | isPathSeparator s1 && isPathSeparator s2 =
         Just (s1:s2:a,b)
     where (a,b) = readDriveShareName xs
-readDriveShare x = Nothing
+readDriveShare _ = Nothing
 
--- assume you have already seen \\
--- share\bob -> "share","\","bob"
+{- assume you have already seen \\ -}
+{- share\bob -> "share","\","bob" -}
 readDriveShareName :: String -> (FilePath, FilePath)
 readDriveShareName name = addSlash a b
     where (a,b) = break isPathSeparator name
@@ -355,7 +361,11 @@ readDriveShareName name = addSlash a b
 
 -- | Join a drive and the rest of the path.
 --
--- > uncurry joinDrive (splitDrive x) == x
+-- >          uncurry joinDrive (splitDrive x) == x
+-- > Windows: joinDrive "C:" "foo" == "C:foo"
+-- > Windows: joinDrive "C:\\" "bar" == "C:\\bar"
+-- > Windows: joinDrive "\\\\share" "foo" == "\\\\share\\foo"
+-- > Windows: joinDrive "/:" "foo" == "/:\\foo"
 joinDrive :: FilePath -> FilePath -> FilePath
 joinDrive a b | isPosix = a ++ b
               | null a = b
@@ -365,12 +375,6 @@ joinDrive a b | isPosix = a ++ b
                                 [a1,':'] | isLetter a1 -> a ++ b
                                 _ -> a ++ [pathSeparator] ++ b
 
--- | Set the drive, from the filepath.
---
--- > replaceDrive x (takeDrive x) == x
-replaceDrive :: FilePath -> String -> FilePath
-replaceDrive x drv = joinDrive drv (dropDrive x)
-
 -- | Get the drive from a filepath.
 --
 -- > takeDrive x == fst (splitDrive x)
@@ -401,7 +405,7 @@ isDrive = null . dropDrive
 -- | Split a filename into directory and file. 'combine' is the inverse.
 --
 -- > uncurry (++) (splitFileName x) == x
--- > uncurry combine (splitFileName (makeValid x)) == (makeValid x)
+-- > Valid x => uncurry combine (splitFileName x) == x
 -- > splitFileName "file/bob.txt" == ("file/", "bob.txt")
 -- > splitFileName "file/" == ("file/", "")
 -- > splitFileName "bob" == ("", "bob")
@@ -416,9 +420,9 @@ splitFileName x = (c ++ reverse b, reverse a)
 
 -- | Set the filename.
 --
--- > replaceFileName (makeValid x) (takeFileName (makeValid x)) == makeValid x
+-- > Valid x => replaceFileName x (takeFileName x) == x
 replaceFileName :: FilePath -> String -> FilePath
-replaceFileName x y = dropFileName x `combine` y
+replaceFileName x y = dropFileName x </> y
 
 -- | Drop the filename.
 --
@@ -430,10 +434,11 @@ dropFileName = fst . splitFileName
 -- | Get the file name.
 --
 -- > takeFileName "test/" == ""
+-- > takeFileName x `isSuffixOf` x
 -- > takeFileName x == snd (splitFileName x)
--- > takeFileName (replaceFileName x "fred") == "fred"
--- > takeFileName (combine x "fred") == "fred"
--- > isRelative (takeFileName (makeValid x))
+-- > Valid x => takeFileName (replaceFileName x "fred") == "fred"
+-- > Valid x => takeFileName (x </> "fred") == "fred"
+-- > Valid x => isRelative (takeFileName x)
 takeFileName :: FilePath -> FilePath
 takeFileName = snd . splitFileName
 
@@ -455,7 +460,7 @@ takeBaseName = dropExtension . takeFileName
 -- > replaceBaseName "/dave/fred/bob.gz.tar" "new" == "/dave/fred/new.tar"
 -- > replaceBaseName x (takeBaseName x) == x
 replaceBaseName :: FilePath -> String -> FilePath
-replaceBaseName pth nam = combineAlways a (addExtension nam ext)
+replaceBaseName pth nam = combineAlways a (nam <.> ext)
     where
         (a,b) = splitFileName pth
         ext = takeExtension b
@@ -472,7 +477,7 @@ hasTrailingPathSeparator x = isPathSeparator (last x)
 -- | Add a trailing file path separator if one is not already present.
 --
 -- > hasTrailingPathSeparator (addTrailingPathSeparator x)
--- > if hasTrailingPathSeparator x then addTrailingPathSeparator x == x else True
+-- > hasTrailingPathSeparator x ==> addTrailingPathSeparator x == x
 -- > Posix:    addTrailingPathSeparator "test/rest" == "test/rest/"
 addTrailingPathSeparator :: FilePath -> FilePath
 addTrailingPathSeparator x = if hasTrailingPathSeparator x then x else x ++ [pathSeparator]
@@ -483,17 +488,22 @@ addTrailingPathSeparator x = if hasTrailingPathSeparator x then x else x ++ [pat
 -- > dropTrailingPathSeparator "file/test/" == "file/test"
 -- > not (hasTrailingPathSeparator (dropTrailingPathSeparator x)) || isDrive x
 -- > Posix:    dropTrailingPathSeparator "/" == "/"
+-- > Windows:  dropTrailingPathSeparator "\\" == "\\"
 dropTrailingPathSeparator :: FilePath -> FilePath
 dropTrailingPathSeparator x =
     if hasTrailingPathSeparator x && not (isDrive x)
-    then reverse $ dropWhile isPathSeparator $ reverse x
+    then let x' = reverse $ dropWhile isPathSeparator $ reverse x
+         in if null x' then [pathSeparator] else x'
     else x
 
 
 -- | Get the directory name, move up one level.
 --
--- > Posix:    takeDirectory "/foo/bar/baz" == "/foo/bar"
--- > Posix:    takeDirectory "/foo/bar/baz/" == "/foo/bar/baz"
+-- >           takeDirectory x `isPrefixOf` x
+-- >           takeDirectory "foo" == ""
+-- >           takeDirectory "/foo/bar/baz" == "/foo/bar"
+-- >           takeDirectory "/foo/bar/baz/" == "/foo/bar/baz"
+-- >           takeDirectory "foo/bar/baz" == "foo/bar"
 -- > Windows:  takeDirectory "foo\\bar" == "foo"
 -- > Windows:  takeDirectory "foo\\bar\\\\" == "foo\\bar"
 -- > Windows:  takeDirectory "C:\\" == "C:\\"
@@ -514,12 +524,13 @@ replaceDirectory x dir = combineAlways dir (takeFileName x)
 
 -- | Combine two paths, if the second path 'isAbsolute', then it returns the second.
 --
--- > combine (takeDirectory (makeValid x)) (takeFileName (makeValid x)) `equalFilePath` makeValid x
+-- > Valid x => combine (takeDirectory x) (takeFileName x) `equalFilePath` x
 -- > Posix:   combine "/" "test" == "/test"
 -- > Posix:   combine "home" "bob" == "home/bob"
 -- > Windows: combine "home" "bob" == "home\\bob"
+-- > Windows: combine "home" "/bob" == "/bob"
 combine :: FilePath -> FilePath -> FilePath
-combine a b | isAbsolute b = b
+combine a b | hasDrive b || (not (null b) && isPathSeparator (head b)) = b
             | otherwise = combineAlways a b
 
 -- | Combine two paths, assuming rhs is NOT absolute.
@@ -545,28 +556,28 @@ combineAlways a b | null a = b
 -- > Windows: splitPath "c:\\test\\path" == ["c:\\","test\\","path"]
 -- > Posix:   splitPath "/file/test" == ["/","file/","test"]
 splitPath :: FilePath -> [FilePath]
-splitPath x = [a | a /= ""] ++ f b
+splitPath x = [drive | drive /= ""] ++ f path
     where
-        (a,b) = splitDrive x
+        (drive,path) = splitDrive x
 
         f "" = []
-        f x = (a++c) : f d
+        f y = (a++c) : f d
             where
-                (a,b) = break isPathSeparator x
+                (a,b) = break isPathSeparator y
                 (c,d) = break (not . isPathSeparator) b
 
 -- | Just as 'splitPath', but don't add the trailing slashes to each element.
 --
 -- > splitDirectories "test/file" == ["test","file"]
 -- > splitDirectories "/test/file" == ["/","test","file"]
--- > joinPath (splitDirectories (makeValid x)) `equalFilePath` makeValid x
+-- > Valid x => joinPath (splitDirectories x) `equalFilePath` x
 -- > splitDirectories "" == []
 splitDirectories :: FilePath -> [FilePath]
-splitDirectories x =
-        if hasDrive x then head xs : f (tail xs)
-        else f xs
+splitDirectories path =
+        if hasDrive path then head pathComponents : f (tail pathComponents)
+        else f pathComponents
     where
-        xs = splitPath x
+        pathComponents = splitPath path
 
         f xs = map g xs
         g x = if null res then x else res
@@ -575,7 +586,9 @@ splitDirectories x =
 
 -- | Join path elements back together.
 --
--- > joinPath (splitPath (makeValid x)) == makeValid x
+-- > Valid x => joinPath (splitPath x) == x
+-- > joinPath [] == ""
+-- > Posix: joinPath ["test","file","path"] == "test/file/path"
 
 -- Note that this definition on c:\\c:\\, join then split will give c:\\.
 joinPath :: [FilePath] -> FilePath
@@ -593,23 +606,36 @@ joinPath x = foldr combine "" x
 --   If you call @System.Directory.canonicalizePath@
 --   first this has a much better chance of working.
 --   Note that this doesn't follow symlinks or DOSNAM~1s.
+--
+-- >          x == y ==> equalFilePath x y
+-- >          normalise x == normalise y ==> equalFilePath x y
+-- > Posix:   equalFilePath "foo" "foo/"
+-- > Posix:   not (equalFilePath "foo" "/foo")
+-- > Posix:   not (equalFilePath "foo" "FOO")
+-- > Windows: equalFilePath "foo" "FOO"
 equalFilePath :: FilePath -> FilePath -> Bool
 equalFilePath a b = f a == f b
     where
         f x | isWindows = dropTrailSlash $ map toLower $ normalise x
             | otherwise = dropTrailSlash $ normalise x
 
-        dropTrailSlash "" = ""
-        dropTrailSlash x | isPathSeparator (last x) = init x
+        dropTrailSlash x | length x >= 2 && isPathSeparator (last x) = init x
                          | otherwise = x
 
 
 -- | Contract a filename, based on a relative path.
 --
--- > Windows: makeRelative x (x `combine` y) == y || takeDrive x == x
--- > Posix:   makeRelative x (x `combine` y) == y
--- > (isRelative x && makeRelative y x == x) || y `combine` makeRelative y x == x
+--   There is no corresponding @makeAbsolute@ function, instead use
+--   @System.Directory.canonicalizePath@ which has the same effect.
+--
+-- >          Valid y => equalFilePath x y || (isRelative x && makeRelative y x == x) || equalFilePath (y </> makeRelative y x) x
+-- >          makeRelative x x == "."
+-- >          null y || equalFilePath (makeRelative x (x </> y)) y || null (takeFileName x)
 -- > Windows: makeRelative "C:\\Home" "c:\\home\\bob" == "bob"
+-- > Windows: makeRelative "C:\\Home" "c:/home/bob" == "bob"
+-- > Windows: makeRelative "C:\\Home" "D:\\Home\\Bob" == "D:\\Home\\Bob"
+-- > Windows: makeRelative "C:\\Home" "C:Home\\Bob" == "C:Home\\Bob"
+-- > Windows: makeRelative "/Home" "/home/bob" == "bob"
 -- > Posix:   makeRelative "/Home" "/home/bob" == "/home/bob"
 -- > Posix:   makeRelative "/home/" "/home/bob/foo/bar" == "bob/foo/bar"
 -- > Posix:   makeRelative "/fred" "bob" == "bob"
@@ -617,24 +643,25 @@ equalFilePath a b = f a == f b
 -- > Posix:   makeRelative "/file/test" "/file/test/fred/" == "fred/"
 -- > Posix:   makeRelative "some/path" "some/path/a/b/c" == "a/b/c"
 makeRelative :: FilePath -> FilePath -> FilePath
-makeRelative x y | not (takeDrive x `equalFilePath` takeDrive y) = y
-makeRelative x orig = f (dropDrive x) (dropDrive orig)
+makeRelative root path
+ | equalFilePath root path = "."
+ | takeAbs root /= takeAbs path = path
+ | otherwise = f (dropAbs root) (dropAbs path)
     where
         f "" y = dropWhile isPathSeparator y
         f x y = let (x1,x2) = g x
                     (y1,y2) = g y
-                in if equalFilePath x1 y1 then f x2 y2 else orig
+                in if equalFilePath x1 y1 then f x2 y2 else path
 
         g x = (dropWhile isPathSeparator a, dropWhile isPathSeparator b)
             where (a,b) = break isPathSeparator $ dropWhile isPathSeparator x
 
+        -- on windows, need to drop '/' which is kind of absolute, but not a drive
+        dropAbs (x:xs) | isPathSeparator x = xs
+        dropAbs x = dropDrive x
 
--- | 'makeRelative' the current directory.
-makeRelativeToCurrentDirectory :: FilePath -> IO FilePath
-makeRelativeToCurrentDirectory x = do
-    cur <- getCurrentDirectory
-    return $ makeRelative cur x
-
+        takeAbs (x:_) | isPathSeparator x = [pathSeparator]
+        takeAbs x = map (\y -> if isPathSeparator y then pathSeparator else toLower y) $ takeDrive x
 
 -- | Normalise a file
 --
@@ -650,66 +677,77 @@ makeRelativeToCurrentDirectory x = do
 -- > Posix:   normalise "../bob/fred/" == "../bob/fred/"
 -- > Posix:   normalise "./bob/fred/" == "bob/fred/"
 -- > Windows: normalise "c:\\file/bob\\" == "C:\\file\\bob\\"
+-- > Windows: normalise "c:\\" == "C:\\"
 -- > Windows: normalise "\\\\server\\test" == "\\\\server\\test"
 -- > Windows: normalise "c:/file" == "C:\\file"
+-- >          normalise "." == "."
+-- > Posix:   normalise "./" == "./"
 normalise :: FilePath -> FilePath
-normalise "" = ""
-normalise x = joinDrive (normaliseDrive drv) (f pth) ++ [pathSeparator | isPathSeparator $ last x]
+normalise path = joinDrive (normaliseDrive drv) (f pth)
+              ++ [pathSeparator | not (null pth) && isPathSeparator (last pth)]
     where
-        (drv,pth) = splitDrive x
+        (drv,pth) = splitDrive path
 
         f = joinPath . dropDots [] . splitDirectories . propSep
 
-        g x = if isPathSeparator x then pathSeparator else x
-
-        propSep (a:b:xs) | isPathSeparator a && isPathSeparator b = propSep (a:xs)
-        propSep (a:xs) | isPathSeparator a = pathSeparator : propSep xs
+        propSep (a:b:xs)
+         | isPathSeparator a && isPathSeparator b = propSep (a:xs)
+        propSep (a:xs)
+         | isPathSeparator a = pathSeparator : propSep xs
         propSep (x:xs) = x : propSep xs
         propSep [] = []
 
-        dropDots acc (".":xs) = dropDots acc xs
+        dropDots acc (".":xs) | not $ null xs = dropDots acc xs
         dropDots acc (x:xs) = dropDots (x:acc) xs
         dropDots acc [] = reverse acc
 
 normaliseDrive :: FilePath -> FilePath
-normaliseDrive x | isPosix = x
-normaliseDrive x = if isJust $ readDriveLetter x2 then
-                       map toUpper x2
-                   else
-                       x
+normaliseDrive drive | isPosix = drive
+normaliseDrive drive = if isJust $ readDriveLetter x2
+                       then map toUpper x2
+                       else drive
     where
-        x2 = map repSlash x
+        x2 = map repSlash drive
 
         repSlash x = if isPathSeparator x then pathSeparator else x
 
 -- information for validity functions on Windows
 -- see http://msdn.microsoft.com/library/default.asp?url=/library/en-us/fileio/fs/naming_a_file.asp
-badCharacters = ":*?><|"
+badCharacters :: [Char]
+badCharacters = ":*?><|\""
+badElements :: [FilePath]
 badElements = ["CON", "PRN", "AUX", "NUL", "COM1", "COM2", "COM3", "COM4", "COM5", "COM6", "COM7", "COM8", "COM9", "LPT1", "LPT2", "LPT3", "LPT4", "LPT5", "LPT6", "LPT7", "LPT8", "LPT9", "CLOCK$"]
 
 
 -- | Is a FilePath valid, i.e. could you create a file like it?
 --
+-- >          isValid "" == False
 -- > Posix:   isValid "/random_ path:*" == True
--- > Posix:   isValid x == True
+-- > Posix:   isValid x == not (null x)
 -- > Windows: isValid "c:\\test" == True
 -- > Windows: isValid "c:\\test:of_test" == False
 -- > Windows: isValid "test*" == False
 -- > Windows: isValid "c:\\test\\nul" == False
 -- > Windows: isValid "c:\\test\\prn.txt" == False
 -- > Windows: isValid "c:\\nul\\file" == False
+-- > Windows: isValid "\\\\" == False
 isValid :: FilePath -> Bool
-isValid x | isPosix = True
-isValid x = not (any (`elem` badCharacters) x2) && not (any f $ splitDirectories x2)
+isValid "" = False
+isValid _ | isPosix = True
+isValid path =
+        not (any (`elem` badCharacters) x2) &&
+        not (any f $ splitDirectories x2) &&
+        not (length path >= 2 && all isPathSeparator path)
     where
-        x2 = dropDrive x
+        x2 = dropDrive path
         f x = map toUpper (dropExtensions x) `elem` badElements
 
 
 -- | Take a FilePath and make it valid; does not change already valid FilePaths.
 --
 -- > isValid (makeValid x)
--- > if isValid x then makeValid x == x else True
+-- > isValid x ==> makeValid x == x
+-- > makeValid "" == "_"
 -- > Windows: makeValid "c:\\test:of_test" == "c:\\test_of_test"
 -- > Windows: makeValid "test*" == "test_"
 -- > Windows: makeValid "c:\\test\\nul" == "c:\\test\\nul_"
@@ -717,10 +755,12 @@ isValid x = not (any (`elem` badCharacters) x2) && not (any f $ splitDirectories
 -- > Windows: makeValid "c:\\test/prn.txt" == "c:\\test/prn_.txt"
 -- > Windows: makeValid "c:\\nul\\file" == "c:\\nul_\\file"
 makeValid :: FilePath -> FilePath
-makeValid x | isPosix = x
-makeValid x = joinDrive drv $ validElements $ validChars pth
+makeValid "" = "_"
+makeValid path | isPosix = path
+makeValid x | length x >= 2 && all isPathSeparator x = take 2 x ++ "drive"
+makeValid path = joinDrive drv $ validElements $ validChars pth
     where
-        (drv,pth) = splitDrive x
+        (drv,pth) = splitDrive path
 
         validChars x = map f x
         f x | x `elem` badCharacters = '_'
@@ -729,7 +769,7 @@ makeValid x = joinDrive drv $ validElements $ validChars pth
         validElements x = joinPath $ map g $ splitPath x
         g x = h (reverse b) ++ reverse a
             where (a,b) = span isPathSeparator $ reverse x
-        h x = if map toUpper a `elem` badElements then addExtension (a ++ "_") b else x
+        h x = if map toUpper a `elem` badElements then a ++ "_" <.> b else x
             where (a,b) = splitExtensions x
 
 
@@ -737,10 +777,25 @@ makeValid x = joinDrive drv $ validElements $ validChars pth
 --
 -- > Windows: isRelative "path\\test" == True
 -- > Windows: isRelative "c:\\test" == False
+-- > Windows: isRelative "c:test" == True
+-- > Windows: isRelative "c:" == True
+-- > Windows: isRelative "\\\\foo" == False
+-- > Windows: isRelative "/foo" == True
 -- > Posix:   isRelative "test/path" == True
 -- > Posix:   isRelative "/test" == False
 isRelative :: FilePath -> Bool
-isRelative = null . takeDrive
+isRelative = isRelativeDrive . takeDrive
+
+
+-- > isRelativeDrive "" == True
+-- > Windows: isRelativeDrive "c:\\" == False
+-- > Windows: isRelativeDrive "c:/" == False
+-- > Windows: isRelativeDrive "c:" == True
+-- > Windows: isRelativeDrive "\\\\foo" == False
+-- > Posix:   isRelativeDrive "/" == False
+isRelativeDrive :: String -> Bool
+isRelativeDrive x = null x ||
+    maybe False (not . isPathSeparator . last . fst) (readDriveLetter x)
 
 
 -- | @not . 'isRelative'@
