@@ -11,30 +11,32 @@
 
 #include <stdio.h>
 
-#if defined(PROFILING) || defined(DEBUG)
-void initProfiling1 ( void );
-void freeProfiling1 ( void );
-void initProfiling2 ( void );
-void endProfiling   ( void );
+BEGIN_RTS_PRIVATE
+
+void initProfiling1 (void);
+void freeProfiling1 (void);
+void initProfiling2 (void);
+void endProfiling   (void);
 
 extern FILE *prof_file;
 extern FILE *hp_file;
-#endif
 
 #ifdef PROFILING
 
-void gen_XML_logfile     ( void );
+void gen_XML_logfile    ( void );
 void reportCCSProfiling ( void );
 
 void PrintNewStackDecls ( void );
 
-extern void fprintCCS( FILE *f, CostCentreStack *ccs );
-extern void fprintCCS_stderr( CostCentreStack *ccs );
+void fprintCCS( FILE *f, CostCentreStack *ccs );
+void fprintCCS_stderr( CostCentreStack *ccs );
 
 #ifdef DEBUG
-extern void debugCCS( CostCentreStack *ccs );
+void debugCCS( CostCentreStack *ccs );
 #endif
 
 #endif
 
+END_RTS_PRIVATE
+
 #endif /* PROFILING_H */
