@@ -16,19 +16,21 @@
 #define INLINE inline
 #endif
 
+#include "PosixSource.h"
 #include "Rts.h"
+
 #include "RtsUtils.h"
 #include "RetainerProfile.h"
 #include "RetainerSet.h"
 #include "Schedule.h"
 #include "Printer.h"
-#include "RtsFlags.h"
 #include "Weak.h"
 #include "Sanity.h"
 #include "Profiling.h"
 #include "Stats.h"
 #include "ProfHeap.h"
 #include "Apply.h"
+#include "sm/Storage.h" // for END_OF_STATIC_LIST
 
 /*
   Note: what to change in order to plug-in a new retainer profiling scheme?
@@ -453,8 +455,6 @@ push( StgClosure *c, retainer c_child_r, StgClosure **first_child )
     case CONSTR_0_2:
     case CAF_BLACKHOLE:
     case BLACKHOLE:
-    case SE_BLACKHOLE:
-    case SE_CAF_BLACKHOLE:
     case ARR_WORDS:
 	*first_child = NULL;
 	return;
@@ -491,7 +491,8 @@ push( StgClosure *c, retainer c_child_r, StgClosure **first_child )
 
 	// three children (fixed), no SRT
 	// need to push a stackElement
-    case MVAR:
+    case MVAR_CLEAN:
+    case MVAR_DIRTY:
 	// head must be TSO and the head of a linked list of TSOs.
 	// Shoule it be a child? Seems to be yes.
 	*first_child = (StgClosure *)((StgMVar *)c)->head;
@@ -620,12 +621,6 @@ push( StgClosure *c, retainer c_child_r, StgClosure **first_child )
     case RET_BIG:
 	// invalid objects
     case IND:
-    case BLOCKED_FETCH:
-    case FETCH_ME:
-    case FETCH_ME_BQ:
-    case RBH:
-    case REMOTE_REF:
-    case EVACUATED:
     case INVALID_OBJECT:
     default:
 	barf("Invalid object *c in push()");
@@ -804,7 +799,8 @@ pop( StgClosure **c, StgClosure **cp, retainer *r )
 
 	    // three children (fixed), no SRT
 	    // need to push a stackElement
-	case MVAR:
+        case MVAR_CLEAN:
+        case MVAR_DIRTY:
 	    if (se->info.next.step == 2) {
 		*c = (StgClosure *)((StgMVar *)se->c)->tail;
 		se->info.next.step++;             // move to the next step
@@ -957,8 +953,6 @@ pop( StgClosure **c, StgClosure **cp, retainer *r )
 	case CONSTR_0_2:
 	case CAF_BLACKHOLE:
 	case BLACKHOLE:
-	case SE_BLACKHOLE:
-	case SE_CAF_BLACKHOLE:
 	case ARR_WORDS:
 	    // one child (fixed), no SRT
 	case MUT_VAR_CLEAN:
@@ -985,12 +979,6 @@ pop( StgClosure **c, StgClosure **cp, retainer *r )
 	case RET_BIG:
 	    // invalid objects
 	case IND:
-	case BLOCKED_FETCH:
-	case FETCH_ME:
-	case FETCH_ME_BQ:
-	case RBH:
-	case REMOTE_REF:
-	case EVACUATED:
 	case INVALID_OBJECT:
 	default:
 	    barf("Invalid object *c in pop()");
@@ -1057,7 +1045,8 @@ isRetainer( StgClosure *c )
     case TSO:
 
 	// mutable objects
-    case MVAR:
+    case MVAR_CLEAN:
+    case MVAR_DIRTY:
     case MUT_VAR_CLEAN:
     case MUT_VAR_DIRTY:
     case MUT_ARR_PTRS_CLEAN:
@@ -1111,8 +1100,6 @@ isRetainer( StgClosure *c )
 	// blackholes
     case CAF_BLACKHOLE:
     case BLACKHOLE:
-    case SE_BLACKHOLE:
-    case SE_CAF_BLACKHOLE:
 	// indirection
     case IND_PERM:
     case IND_OLDGEN_PERM:
@@ -1149,12 +1136,6 @@ isRetainer( StgClosure *c )
     case RET_BIG:
 	// other cases
     case IND:
-    case BLOCKED_FETCH:
-    case FETCH_ME:
-    case FETCH_ME_BQ:
-    case RBH:
-    case REMOTE_REF:
-    case EVACUATED:
     case INVALID_OBJECT:
     default:
 	barf("Invalid object in isRetainer(): %d", get_itbl(c)->type);
@@ -1632,7 +1613,7 @@ inner_loop:
 #ifdef DEBUG_RETAINER
 	    debugBelch("ThreadRelocated encountered in retainClosure()\n");
 #endif
-	    c = (StgClosure *)((StgTSO *)c)->link;
+	    c = (StgClosure *)((StgTSO *)c)->_link;
 	    goto inner_loop;
 	}
 	break;
@@ -1797,7 +1778,7 @@ inner_loop:
  *  Compute the retainer set for every object reachable from *tl.
  * -------------------------------------------------------------------------- */
 static void
-retainRoot( StgClosure **tl )
+retainRoot(void *user STG_UNUSED, StgClosure **tl)
 {
     StgClosure *c;
 
@@ -1834,7 +1815,7 @@ computeRetainerSet( void )
     RetainerSet tmpRetainerSet;
 #endif
 
-    GetRoots(retainRoot);	// for scheduler roots
+    markCapabilities(retainRoot, NULL);	// for scheduler roots
 
     // This function is called after a major GC, when key, value, and finalizer
     // all are guaranteed to be valid, or reachable.
@@ -1843,10 +1824,10 @@ computeRetainerSet( void )
     // for retainer profilng.
     for (weak = weak_ptr_list; weak != NULL; weak = weak->link)
 	// retainRoot((StgClosure *)weak);
-	retainRoot((StgClosure **)&weak);
+	retainRoot(NULL, (StgClosure **)&weak);
 
     // Consider roots from the stable ptr table.
-    markStablePtrTable(retainRoot);
+    markStablePtrTable(retainRoot, NULL);
 
     // The following code resets the rs field of each unvisited mutable
     // object (computing sumOfNewCostExtra and updating costArray[] when
@@ -1910,7 +1891,7 @@ computeRetainerSet( void )
  *    they are not taken into consideration in computing retainer sets.
  * -------------------------------------------------------------------------- */
 void
-resetStaticObjectForRetainerProfiling( void )
+resetStaticObjectForRetainerProfiling( StgClosure *static_objects )
 {
 #ifdef DEBUG_RETAINER
     nat count;
@@ -1920,7 +1901,7 @@ resetStaticObjectForRetainerProfiling( void )
 #ifdef DEBUG_RETAINER
     count = 0;
 #endif
-    p = scavenged_static_objects;
+    p = static_objects;
     while (p != END_OF_STATIC_LIST) {
 #ifdef DEBUG_RETAINER
 	count++;
@@ -2167,7 +2148,7 @@ smallObjectPoolCheck(void)
     StgPtr p;
     static nat costSum, size;
 
-    bd = small_alloc_list;
+    bd = g0s0->blocks;
     costSum = 0;
 
     // first block
