@@ -11,12 +11,15 @@
 
 #include "GetTime.h"
 
+BEGIN_RTS_PRIVATE
+
 void      stat_startInit(void);
 void      stat_endInit(void);
 
 void      stat_startGC(void);
 void      stat_endGC (lnat alloc, lnat live, 
-		      lnat copied, lnat scavd_copied, lnat gen);
+		      lnat copied, lnat gen,
+                      lnat max_copied, lnat avg_copied, lnat slop);
 
 #ifdef PROFILING
 void      stat_startRP(void);
@@ -38,7 +41,8 @@ void      stat_endExit(void);
 void      stat_exit(int alloc);
 void      stat_workerStop(void);
 
-void      initStats(void);
+void      initStats0(void);
+void      initStats1(void);
 
 double    mut_user_time_during_GC(void);
 double    mut_user_time(void);
@@ -49,7 +53,6 @@ double    mut_user_time_during_heap_census(void);
 #endif /* PROFILING */
 
 void      statDescribeGens( void );
-HsInt64   getAllocations( void );
 
 Ticks stat_getElapsedGCTime(void);
 Ticks stat_getElapsedTime(void);
@@ -58,5 +61,6 @@ Ticks stat_getElapsedTime(void);
 void statsPrintf( char *s, ... ) 
     GNUC3_ATTRIBUTE(format (printf, 1, 2));
 
+END_RTS_PRIVATE
 
 #endif /* STATS_H */
