@@ -14,12 +14,13 @@
  * on platform-specific services to install and run the timers.
  *
  */
+
+#include "PosixSource.h"
 #include "Rts.h"
-#include "RtsFlags.h"
-#include "Proftimer.h"
-#include "Storage.h"
-#include "Schedule.h"
+
 #include "Timer.h"
+#include "Proftimer.h"
+#include "Schedule.h"
 #include "Ticker.h"
 #include "Capability.h"
 #include "RtsSignals.h"
@@ -47,7 +48,7 @@ handle_tick(int unused STG_UNUSED)
       ticks_to_ctxt_switch--;
       if (ticks_to_ctxt_switch <= 0) {
 	  ticks_to_ctxt_switch = RtsFlags.ConcFlags.ctxtSwitchTicks;
-	  context_switch = 1;	/* schedule a context switch */
+	  setContextSwitches();	/* schedule a context switch */
       }
   }
 
