@@ -9,16 +9,20 @@
 #ifndef BLOCK_ALLOC_H
 #define BLOCK_ALLOC_H
 
+BEGIN_RTS_PRIVATE
+
 /* Debugging  -------------------------------------------------------------- */
 
 #ifdef DEBUG
-extern void checkFreeListSanity(void);
-nat         countFreeList(void);
-void        markBlocks (bdescr *bd);
-void        reportUnmarkedBlocks (void);
+void checkFreeListSanity(void);
+nat  countFreeList(void);
+void markBlocks (bdescr *bd);
+void reportUnmarkedBlocks (void);
 #endif
 
 extern lnat n_alloc_blocks;   // currently allocated blocks
 extern lnat hw_alloc_blocks;  // high-water allocated blocks
 
+END_RTS_PRIVATE
+
 #endif /* BLOCK_ALLOC_H */
