@@ -11,8 +11,10 @@
  *
  * ---------------------------------------------------------------------------*/
 
-#ifndef GCCOMPACT_H
-#define GCCOMPACT_H
+#ifndef SM_COMPACT_H
+#define SM_COMPACT_H
+
+BEGIN_RTS_PRIVATE
 
 INLINE_HEADER rtsBool
 mark_stack_empty(void)
@@ -74,6 +76,8 @@ is_marked(StgPtr p, bdescr *bd)
     return (*bitmap_word & bit_mask);
 }
 
-extern void compact (StgClosure *static_objects);
+void compact (StgClosure *static_objects);
 
-#endif /* GCCOMPACT_H */
+END_RTS_PRIVATE
+
+#endif /* SM_COMPACT_H */
